from __future__ import annotations

from typing import Tuple, Dict, TYPE_CHECKING

import numpy
from amulet_nbt import (
    NBTFile,
    TAG_Byte,
    TAG_Int,
    TAG_Long,
    TAG_String,
    TAG_List,
    TAG_Compound,
    TAG_Byte_Array,
    TAG_Int_Array,
)

import amulet
from amulet.api.data_types import SubChunkNDArray, AnyNDArray, BlockCoordinates
from amulet.utils import world_utils
from amulet.api.wrapper import EntityIDType, EntityCoordType
from amulet.api.chunk import Chunk, StatusFormats
from .base_anvil_interface import (
    BaseAnvilInterface,
)

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk


class AnvilNAInterface(BaseAnvilInterface):
    BlockEntities = "TileEntities"
    Entities = "Entities"
    Sections = "Sections"

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "256IA")

        self._set_feature("light_optional", "false")

        self._set_feature("block_entity_format", EntityIDType.namespace_str_id)
        self._set_feature("block_entity_coord_format", EntityCoordType.xyz_int)

        self._set_feature("entity_format", EntityIDType.namespace_str_id)
        self._set_feature("entity_coord_format", EntityCoordType.Pos_list_double)

    @staticmethod
    def minor_is_valid(key: int):
        return key == -1

    def decode(
        self, cx: int, cz: int, nbt_file: NBTFile, bounds: Tuple[int, int]
    ) -> Tuple["Chunk", AnyNDArray]:
        chunk, root = self._init_decode(cx, cz, nbt_file)
        return self._decode_chunk(chunk, root, bounds)

    @staticmethod
    def _init_decode(cx: int, cz: int, data: NBTFile) -> Tuple[Chunk, TAG_Compound]:
        """Get the decode started by creating a chunk object."""
        chunk = Chunk(cx, cz)
        assert isinstance(data.value, TAG_Compound), "Raw data must be a compound."
        chunk.misc = {
            # store the chunk data so that any non-versioned data can get saved back
            "java_chunk_data": data.value
        }
        return chunk, data.value

    def _decode_chunk(
        self, chunk: Chunk, root: TAG_Compound, bounds: Tuple[int, int]
    ) -> Tuple["Chunk", AnyNDArray]:
        self._decode_root(chunk, root)
        assert self.check_type(root, "Level", TAG_Compound)
        level = root.pop("Level")
        self._decode_level(chunk, level, bounds, bounds[0] >> 4)
        sections = self._extract_sections(chunk, level)
        self._decode_sections(chunk, sections)
        palette = self._decode_blocks(chunk, sections)
        return chunk, palette

    def _decode_root(self, chunk: Chunk, root: TAG_Compound):
        pass

    def _decode_level(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int], floor_cy: int
    ):
        self._decode_coords(chunk, level)
        self._decode_last_update(chunk, level)
        self._decode_status(chunk, level)
        self._decode_inhabited_time(chunk, level)
        self._decode_biomes(chunk, level, floor_cy)
        self._decode_height(chunk, level, bounds)
        self._decode_entities(chunk, level)
        self._decode_block_entities(chunk, level)
        self._decode_block_ticks(chunk, level, floor_cy)

    def _decode_sections(self, chunk: Chunk, sections: Dict[int, TAG_Compound]):
        self._decode_block_light(chunk, sections)
        self._decode_sky_light(chunk, sections)

    @staticmethod
    def _decode_coords(chunk: Chunk, level: TAG_Compound):
        assert chunk.coordinates == (level.pop("xPos"), level.pop("zPos"))

    def _decode_last_update(self, chunk: Chunk, compound: TAG_Compound):
        chunk.misc["last_update"] = self.get_obj(compound, "LastUpdate", TAG_Long).value

    def _decode_status(self, chunk: Chunk, compound: TAG_Compound):
        status = "empty"
        if self.get_obj(compound, "TerrainPopulated", TAG_Byte):
            status = "decorated"
        if self.get_obj(compound, "LightPopulated", TAG_Byte):
            status = "postprocessed"
        chunk.status = status

    def _decode_v_tag(self, chunk: Chunk, compound: TAG_Compound):
        chunk.misc["V"] = self.get_obj(compound, "V", TAG_Byte, TAG_Byte(1)).value

    def _decode_inhabited_time(self, chunk: Chunk, compound: TAG_Compound):
        chunk.misc["inhabited_time"] = self.get_obj(
            compound, "InhabitedTime", TAG_Long
        ).value

    def _decode_biomes(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        biomes = compound.pop("Biomes", None)
        if isinstance(biomes, TAG_Byte_Array) and biomes.value.size == 256:
            chunk.biomes = biomes.astype(numpy.uint32).reshape((16, 16))

    def _decode_height(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        height = self.get_obj(compound, "HeightMap", TAG_Int_Array).value
        if isinstance(height, numpy.ndarray) and height.size == 256:
            chunk.misc["height_map256IA"] = height.reshape((16, 16))

    def _extract_sections(
        self, chunk: Chunk, compound: TAG_Compound
    ) -> Dict[int, TAG_Compound]:
        # parse sections into a more usable format
        sections: Dict[int, TAG_Compound] = {
            section.pop("Y").value: section
            for section in self.get_obj(compound, self.Sections, TAG_List)
            if isinstance(section, TAG_Compound)
            and self.check_type(section, "Y", TAG_Byte)
        }
        chunk.misc["java_sections"] = sections
        return sections

    def _decode_blocks(
        self, chunk: Chunk, chunk_sections: Dict[int, TAG_Compound]
    ) -> AnyNDArray:
        blocks: Dict[int, SubChunkNDArray] = {}
        palette = []
        palette_len = 0
        for cy, section in chunk_sections.items():
            section_blocks = numpy.frombuffer(
                section.pop("Blocks").value, dtype=numpy.uint8
            )
            section_data = numpy.frombuffer(
                section.pop("Data").value, dtype=numpy.uint8
            )
            section_blocks = section_blocks.reshape((16, 16, 16))
            section_blocks = section_blocks.astype(numpy.uint16)

            section_data = world_utils.from_nibble_array(section_data)
            section_data = section_data.reshape((16, 16, 16))

            if "Add" in section:
                add_blocks = numpy.frombuffer(
                    section.pop("Add").value, dtype=numpy.uint8
                )
                add_blocks = world_utils.from_nibble_array(add_blocks)
                add_blocks = add_blocks.reshape((16, 16, 16))

                section_blocks |= add_blocks.astype(numpy.uint16) << 8
                # TODO: fix this

            (section_palette, blocks[cy]) = world_utils.fast_unique(
                numpy.transpose(
                    (section_blocks << 4) + section_data, (2, 0, 1)
                )  # YZX -> XYZ
            )
            blocks[cy] += palette_len
            palette_len += len(section_palette)
            palette.append(section_palette)

        if palette:
            final_palette, lut = numpy.unique(
                numpy.concatenate(palette), return_inverse=True
            )
            final_palette: numpy.ndarray = numpy.array(
                [final_palette >> 4, final_palette & 15]
            ).T
            for cy in blocks:
                blocks[cy] = lut[blocks[cy]]
        else:
            final_palette = numpy.array([], dtype=object)
        chunk.blocks = blocks
        return final_palette

    def _unpack_light(
        self, sections: Dict[int, TAG_Compound], section_key: str
    ) -> Dict[int, numpy.ndarray]:
        light_container = {}
        for cy, section in sections.items():
            if self.check_type(section, section_key, TAG_Byte_Array):
                light: numpy.ndarray = section.pop(section_key).value
                if light.size == 2048:
                    # TODO: check if this needs transposing or if the values are the other way around
                    light_container[cy] = (
                        (
                            light.reshape(-1, 1)
                            & numpy.array([0xF, 0xF0], dtype=numpy.uint8)
                        )
                        >> numpy.array([0, 4], dtype=numpy.uint8)
                    ).reshape((16, 16, 16))
        return light_container

    def _decode_block_light(
        self, chunk: Chunk, chunk_sections: Dict[int, TAG_Compound]
    ):
        chunk.misc["block_light"] = self._unpack_light(chunk_sections, "BlockLight")

    def _decode_sky_light(self, chunk: Chunk, chunk_sections: Dict[int, TAG_Compound]):
        chunk.misc["sky_light"] = self._unpack_light(chunk_sections, "SkyLight")

    def _decode_entities(self, chunk: Chunk, compound: TAG_Compound):
        ents = self._decode_entity_list(self.get_obj(compound, self.Entities, TAG_List))
        if amulet.entity_support:
            chunk.entities = ents
        else:
            chunk.misc["java_entities_temp"] = ents

    def _decode_block_entities(self, chunk: Chunk, compound: TAG_Compound):
        chunk.block_entities = self._decode_block_entity_list(
            self.get_obj(compound, self.BlockEntities, TAG_List)
        )

    @staticmethod
    def _decode_ticks(ticks: TAG_List) -> Dict[BlockCoordinates, Tuple[str, int, int]]:
        return {
            (tick["x"].value, tick["y"].value, tick["z"].value): (
                tick["i"].value,
                tick["t"].value,
                tick["p"].value,
            )
            for tick in ticks
            if all(c in tick and isinstance(tick[c], TAG_Int) for c in "xyztp")
            and "i" in tick
            and isinstance(tick["i"], TAG_String)
        }

    def _decode_block_ticks(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        chunk.misc.setdefault("block_ticks", {}).update(
            self._decode_ticks(self.get_obj(compound, "TileTicks", TAG_List))
        )

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, int],
        bounds: Tuple[int, int],
    ) -> NBTFile:
        return NBTFile(self._encode_chunk(chunk, palette, max_world_version, bounds))

    def _encode_chunk(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, int],
        bounds: Tuple[int, int],
    ) -> TAG_Compound:
        root = self._init_encode(chunk)
        self._encode_root(root, max_world_version)
        level: TAG_Compound = self.set_obj(root, "Level", TAG_Compound)
        self._encode_level(chunk, level, bounds)
        sections = self._init_sections(chunk)
        self._encode_sections(chunk, sections, bounds)
        self._encode_blocks(chunk, sections, palette, bounds)
        # these data versions probably extend a little into the snapshots as well
        if 1519 <= max_world_version[1] <= 1631:
            # Java 1.13 to 1.13.2 cannot have empty sections
            for cy in list(sections.keys()):
                if "BlockStates" not in sections[cy] or "Palette" not in sections[cy]:
                    del sections[cy]
        sections_list = []
        for cy, section in sections.items():
            section["Y"] = TAG_Byte(cy)
            sections_list.append(section)
        level[self.Sections] = TAG_List(sections_list)
        return root

    @staticmethod
    def _init_encode(chunk: "Chunk"):
        """Get or create the root tag."""
        if "java_chunk_data" in chunk.misc and isinstance(
            chunk.misc["java_chunk_data"], TAG_Compound
        ):
            return chunk.misc["java_chunk_data"]
        else:
            return TAG_Compound()

    def _encode_root(self, root: TAG_Compound, max_world_version: Tuple[str, int]):
        self._encode_data_version(root, max_world_version)

    def _encode_data_version(
        self, root: TAG_Compound, max_world_version: Tuple[str, int]
    ):
        if "DataVersion" in root:
            del root["DataVersion"]

    def _encode_level(self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]):
        self._encode_coords(chunk, level, bounds)
        self._encode_last_update(chunk, level)
        self._encode_status(chunk, level)
        self._encode_inhabited_time(chunk, level)
        self._encode_biomes(chunk, level, bounds)
        self._encode_height(chunk, level, bounds)
        self._encode_entities(chunk, level)
        self._encode_block_entities(chunk, level)
        self._encode_block_ticks(chunk, level, bounds)

    @staticmethod
    def _init_sections(chunk: "Chunk") -> Dict[int, TAG_Compound]:
        """Get or create the root tag."""
        if "java_sections" in chunk.misc and isinstance(
            chunk.misc["java_sections"], dict
        ):
            # verify that the data is correctly formatted.
            return {
                cy: section
                for cy, section in chunk.misc["java_sections"].items()
                if isinstance(cy, int) and isinstance(section, TAG_Compound)
            }
        else:
            return {}

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        palette: AnyNDArray,
        cy: int,
    ) -> bool:
        if cy in chunk.blocks:
            block_sub_array = palette[
                numpy.transpose(
                    chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
                ).ravel()  # XYZ -> YZX
            ]

            data_sub_array = block_sub_array[:, 1]
            block_sub_array = block_sub_array[:, 0]
            # if not numpy.any(block_sub_array) and not numpy.any(data_sub_array):
            #     return False
            section = sections.setdefault(cy, TAG_Compound())
            section["Blocks"] = TAG_Byte_Array(block_sub_array.astype("uint8"))
            section["Data"] = TAG_Byte_Array(
                world_utils.to_nibble_array(data_sub_array)
            )
            return True
        return False

    def _encode_blocks(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        palette: AnyNDArray,
        bounds: Tuple[int, int],
    ):
        for cy in range(bounds[0] >> 4, bounds[1] >> 4):
            if not self._encode_block_section(chunk, sections, palette, cy):
                # In 1.13.2 and before if the Blocks key does not exist but the sub-chunk TAG_Compound does
                # exist the game will error and recreate the chunk. All sub-chunks that do not contain data
                # must be deleted.
                if cy in sections:
                    del sections[cy]

    @staticmethod
    def _encode_coords(chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]):
        level["xPos"] = TAG_Int(chunk.cx)
        level["zPos"] = TAG_Int(chunk.cz)

    @staticmethod
    def _encode_last_update(chunk: Chunk, level: TAG_Compound):
        level["LastUpdate"] = TAG_Long(chunk.misc.get("last_update", 0))

    def _encode_status(self, chunk: Chunk, level: TAG_Compound):
        status = chunk.status.as_type(StatusFormats.Raw)
        level["TerrainPopulated"] = TAG_Byte(int(status > -0.3))
        level["LightPopulated"] = TAG_Byte(int(status > -0.2))

    @staticmethod
    def _encode_v_tag(chunk: Chunk, level: TAG_Compound):
        level["V"] = TAG_Byte(chunk.misc.get("V", 1))

    def _encode_inhabited_time(self, chunk: Chunk, level: TAG_Compound):
        level["InhabitedTime"] = TAG_Long(chunk.misc.get("inhabited_time", 0))

    def _encode_biomes(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        chunk.biomes.convert_to_2d()
        level["Biomes"] = TAG_Byte_Array(chunk.biomes.astype(dtype=numpy.uint8).ravel())

    def _encode_height(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        height = chunk.misc.get("height_map256IA", None)
        if (
            isinstance(height, numpy.ndarray)
            and numpy.issubdtype(height.dtype, numpy.integer)
            and height.shape == (16, 16)
        ):
            level["HeightMap"] = TAG_Int_Array(height.ravel())
        elif self._features["height_map"] == "256IARequired":
            level["HeightMap"] = TAG_Int_Array(numpy.zeros(256, dtype=numpy.uint32))

    def _encode_entities(self, chunk: Chunk, level: TAG_Compound):
        if amulet.entity_support:
            level[self.Entities] = self._encode_entity_list(chunk.entities)
        else:
            level[self.Entities] = self._encode_entity_list(
                chunk.misc.get("java_entities_temp", TAG_List())
            )

    def _encode_block_entities(self, chunk: Chunk, level: TAG_Compound):
        level[self.BlockEntities] = self._encode_block_entity_list(chunk.block_entities)

    @staticmethod
    def _encode_ticks(ticks: Dict[BlockCoordinates, Tuple[str, int, int]]) -> TAG_List:
        ticks_out = TAG_List()
        if isinstance(ticks, dict):
            for k, v in ticks.items():
                try:
                    (x, y, z), (i, t, p) = k, v
                    ticks_out.append(
                        TAG_Compound(
                            {
                                "i": TAG_String(i),
                                "p": TAG_Int(p),
                                "t": TAG_Int(t),
                                "x": TAG_Int(x),
                                "y": TAG_Int(y),
                                "z": TAG_Int(x),
                            }
                        )
                    )
                except Exception:
                    amulet.log.error(f"Could not serialise tick data {k}: {v}")
        return ticks_out

    def _encode_block_ticks(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        level["TileTicks"] = self._encode_ticks(chunk.misc.get("block_ticks", {}))

    def _encode_sections(
        self, chunk: Chunk, sections: Dict[int, TAG_Compound], bounds: Tuple[int, int]
    ):
        self._encode_block_light(chunk, sections)
        self._encode_sky_light(chunk, sections)

    def _pack_light(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        feature_key: str,
        section_key: str,
    ):
        light_container = chunk.misc.get(feature_key, {})
        if not isinstance(light_container, dict):
            light_container = {}
        for cy, section in sections.items():
            light = light_container.get(cy, None)
            if (
                isinstance(light, numpy.ndarray)
                and numpy.issubdtype(light.dtype, numpy.integer)
                and light.shape == (16, 16, 16)
            ):
                light = light.ravel() % 16
                section[section_key] = TAG_Byte_Array(light[::2] + (light[1::2] << 4))
            elif self._features["light_optional"] == "false":
                section[section_key] = TAG_Byte_Array(
                    numpy.full(2048, 255, dtype=numpy.uint8)
                )

    def _encode_block_light(self, chunk: Chunk, sections: Dict[int, TAG_Compound]):
        self._pack_light(chunk, sections, "block_light", "BlockLight")

    def _encode_sky_light(self, chunk: Chunk, sections: Dict[int, TAG_Compound]):
        self._pack_light(chunk, sections, "sky_light", "SkyLight")


export = AnvilNAInterface



from __future__ import annotations
import struct

from .leveldb_28 import (
    LevelDB28Interface as ParentInterface,
)


class LevelDB29Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 29)

    @staticmethod
    def _get_sub_chunk_storage_byte(cy: int, min_y: int) -> bytes:
        return struct.pack("b", cy)


export = LevelDB29Interface



import unittest
from amulet.api.block import Block
from amulet.api.registry import BlockManager
from amulet.api.errors import BlockException

import amulet_nbt as nbt


class BlockTestCase(unittest.TestCase):
    def test_get_from_blockstate(self):  # This is mostly just sanity checks
        air = Block.from_string_blockstate("minecraft:air")

        self.assertIsInstance(air, Block)
        self.assertEqual("minecraft", air.namespace)
        self.assertEqual("air", air.base_name)
        self.assertEqual({}, air.properties)
        self.assertEqual((), air.extra_blocks)
        self.assertEqual("minecraft:air", air.blockstate)
        self.assertEqual("minecraft:air", air.snbt_blockstate)
        self.assertEqual("minecraft:air", air.full_blockstate)

        stone = Block.from_string_blockstate("minecraft:stone")

        self.assertIsInstance(stone, Block)
        self.assertEqual("minecraft", stone.namespace)
        self.assertEqual("stone", stone.base_name)
        self.assertEqual({}, stone.properties)
        self.assertEqual((), stone.extra_blocks)
        self.assertEqual("minecraft:stone", stone.blockstate)
        self.assertEqual("minecraft:stone", stone.snbt_blockstate)
        self.assertEqual("minecraft:stone", stone.full_blockstate)

        self.assertNotEqual(air, stone)

        oak_leaves = Block.from_snbt_blockstate(
            'minecraft:oak_leaves[distance="1",persistent="true"]'
        )

        self.assertIsInstance(oak_leaves, Block)
        self.assertEqual("minecraft", oak_leaves.namespace)
        self.assertEqual("oak_leaves", oak_leaves.base_name)
        self.assertEqual(
            {"distance": nbt.TAG_String("1"), "persistent": nbt.TAG_String("true")},
            oak_leaves.properties,
        )
        self.assertEqual((), oak_leaves.extra_blocks)
        self.assertEqual(
            "minecraft:oak_leaves[distance=1,persistent=true]",
            oak_leaves.blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves.snbt_blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves.full_blockstate,
        )

        oak_leaves_2 = Block(
            namespace="minecraft",
            base_name="oak_leaves",
            properties={
                "persistent": nbt.TAG_String("true"),
                "distance": nbt.TAG_String("1"),
            },
        )
        self.assertEqual("minecraft", oak_leaves_2.namespace)
        self.assertEqual("oak_leaves", oak_leaves_2.base_name)
        self.assertEqual(
            {"distance": nbt.TAG_String("1"), "persistent": nbt.TAG_String("true")},
            oak_leaves_2.properties,
        )
        self.assertEqual((), oak_leaves_2.extra_blocks)
        self.assertEqual(
            "minecraft:oak_leaves[distance=1,persistent=true]",
            oak_leaves_2.blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves_2.snbt_blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves_2.full_blockstate,
        )
        self.assertEqual(oak_leaves, oak_leaves_2)

        oak_leaves_3 = Block.from_snbt_blockstate(
            'minecraft:oak_leaves[persistent="true",distance="1"]'
        )
        self.assertEqual("minecraft", oak_leaves_3.namespace)
        self.assertEqual("oak_leaves", oak_leaves_3.base_name)
        self.assertEqual(
            {"distance": nbt.TAG_String("1"), "persistent": nbt.TAG_String("true")},
            oak_leaves_3.properties,
        )
        self.assertEqual((), oak_leaves_3.extra_blocks)
        self.assertEqual(
            "minecraft:oak_leaves[distance=1,persistent=true]",
            oak_leaves_3.blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves_3.snbt_blockstate,
        )
        self.assertEqual(
            'minecraft:oak_leaves[distance="1",persistent="true"]',
            oak_leaves_3.full_blockstate,
        )
        self.assertEqual(oak_leaves, oak_leaves_3)

    def test_extra_blocks(self):
        stone = Block.from_string_blockstate("minecraft:stone")
        water = Block.from_string_blockstate("minecraft:water[level=1]")
        granite = Block.from_string_blockstate("minecraft:granite")
        dirt = Block.from_string_blockstate("minecraft:dirt")

        conglomerate_1 = stone + water + dirt
        self.assertIsInstance(conglomerate_1, Block)
        self.assertEqual("minecraft", conglomerate_1.namespace)
        self.assertEqual("stone", conglomerate_1.base_name)
        self.assertEqual({}, conglomerate_1.properties)
        self.assertEqual(2, len(conglomerate_1.extra_blocks))
        for block_1, block_2 in zip(conglomerate_1.extra_blocks, (water, dirt)):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        conglomerate_2 = conglomerate_1 + granite
        self.assertIsInstance(conglomerate_2, Block)
        self.assertEqual("minecraft", conglomerate_2.namespace)
        self.assertEqual("stone", conglomerate_2.base_name)
        self.assertEqual({}, conglomerate_2.properties)
        self.assertEqual(3, len(conglomerate_2.extra_blocks))
        for block_1, block_2 in zip(
            conglomerate_2.extra_blocks, (water, dirt, granite)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        self.assertNotEqual(conglomerate_1, conglomerate_2)

        conglomerate_3 = conglomerate_2 - dirt
        self.assertIsInstance(conglomerate_3, Block)
        self.assertEqual("minecraft", conglomerate_3.namespace)
        self.assertEqual("stone", conglomerate_3.base_name)
        self.assertEqual({}, conglomerate_3.properties)
        self.assertEqual(2, len(conglomerate_3.extra_blocks))
        for block_1, block_2 in zip(conglomerate_3.extra_blocks, (water, granite)):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        self.assertRaises(TypeError, lambda: stone + 1)
        self.assertRaises(TypeError, lambda: stone - 1)

    def test_extra_blocks_immutable(self):
        stone = Block.from_string_blockstate("minecraft:stone")
        dirt = Block.from_string_blockstate("minecraft:dirt")

        stone2 = stone
        self.assertIs(stone, stone2)
        stone2 += dirt
        self.assertIsNot(stone, stone2)

        stone3 = stone2
        self.assertIs(stone2, stone3)
        stone3 -= dirt
        self.assertIsNot(stone, stone3)

    def test_remove_layer(self):
        stone = Block.from_string_blockstate("minecraft:stone")
        water = Block.from_string_blockstate("minecraft:water[level=1]")
        granite = Block.from_string_blockstate("minecraft:granite")
        dirt = Block.from_string_blockstate("minecraft:dirt")
        oak_log_axis_x = Block.from_string_blockstate("minecraft:oak_log[axis=x]")

        conglomerate_1 = stone + water + dirt + dirt + granite
        self.assertIsInstance(conglomerate_1, Block)
        self.assertEqual("minecraft", conglomerate_1.namespace)
        self.assertEqual("stone", conglomerate_1.base_name)
        self.assertEqual({}, conglomerate_1.properties)
        self.assertEqual(4, len(conglomerate_1.extra_blocks))
        for block_1, block_2 in zip(
            conglomerate_1.extra_blocks, (water, dirt, dirt, granite)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        new_conglomerate = conglomerate_1.remove_layer(2)
        for block_1, block_2 in zip(
            new_conglomerate.extra_blocks, (water, dirt, granite)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        conglomerate_2 = granite + water + stone + dirt + oak_log_axis_x
        self.assertIsInstance(conglomerate_2, Block)
        self.assertEqual("minecraft", conglomerate_2.namespace)
        self.assertEqual("granite", conglomerate_2.base_name)
        self.assertEqual({}, conglomerate_2.properties)
        self.assertEqual(4, len(conglomerate_2.extra_blocks))
        for block_1, block_2 in zip(
            conglomerate_2.extra_blocks, (water, stone, dirt, oak_log_axis_x)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        new_conglomerate = conglomerate_2.remove_layer(3)
        self.assertEqual(3, len(new_conglomerate.extra_blocks))
        for block_1, block_2 in zip(
            new_conglomerate.extra_blocks, (water, stone, oak_log_axis_x)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        new_base = conglomerate_2.remove_layer(0)
        self.assertEqual(3, len(new_base.extra_blocks))
        self.assertEqual("minecraft", new_base.namespace)
        self.assertEqual("water", new_base.base_name)
        self.assertEqual({"level": nbt.TAG_String("1")}, new_base.properties)
        for block_1, block_2 in zip(
            new_base.extra_blocks, (stone, dirt, oak_log_axis_x)
        ):
            self.assertEqual(block_1, block_2)
            self.assertEqual(0, len(block_1.extra_blocks))

        self.assertNotEqual(new_base, new_base.remove_layer(1))

        with self.assertRaises(BlockException):
            no_block = granite.remove_layer(0)

        with self.assertRaises(BlockException):
            non_present_layer = granite.remove_layer(7)
            non_present_layer = conglomerate_2.remove_layer(5)

        conglomerate_2.remove_layer(4)  # Check if last layer can still be removed

    def test_hash(self):
        stone = Block.from_string_blockstate("minecraft:stone")
        water = Block.from_string_blockstate("minecraft:water[level=1]")
        granite = Block.from_string_blockstate("minecraft:granite")
        dirt = Block.from_string_blockstate("minecraft:dirt")

        conglomerate_1 = stone + water + dirt
        conglomerate_2 = stone + dirt + water

        self.assertNotEqual(conglomerate_1, conglomerate_2)
        self.assertNotEqual(hash(conglomerate_1), hash(conglomerate_2))

        conglomerate_3 = dirt + water + dirt
        conglomerate_4 = dirt + dirt + water

        self.assertNotEqual(conglomerate_3, conglomerate_4)
        self.assertNotEqual(hash(conglomerate_3), hash(conglomerate_4))

        conglomerate_5 = conglomerate_1 + granite
        conglomerate_6 = conglomerate_3 + granite

        self.assertNotEqual(conglomerate_1, conglomerate_5)
        self.assertNotEqual(conglomerate_3, conglomerate_6)
        self.assertNotEqual(conglomerate_5, conglomerate_6)
        self.assertNotEqual(hash(conglomerate_5), hash(conglomerate_6))


class BlockManagerTestCase(unittest.TestCase):
    def setUp(self):
        self.manager = BlockManager()

        initial_dirt = Block.from_string_blockstate("minecraft:dirt")
        initial_stone = Block.from_string_blockstate("minecraft:stone")
        initial_granite = Block.from_string_blockstate("minecraft:granite")

        initial_dirt_water = initial_dirt + Block.from_string_blockstate(
            "minecraft:water"
        )

        # Partially populate the manager
        self.manager.get_add_block(initial_dirt)
        self.manager.get_add_block(initial_stone)
        self.manager.get_add_block(initial_granite)
        self.manager.get_add_block(initial_dirt_water)

    def test_get_index_from_manager(self):
        dirt = Block.from_string_blockstate("minecraft:dirt")
        stone = Block.from_string_blockstate("minecraft:stone")
        granite = Block.from_string_blockstate("minecraft:granite")

        self.assertEqual(0, self.manager[dirt])
        self.assertEqual(1, self.manager[stone])
        self.assertEqual(2, self.manager[granite])

        water = Block.from_string_blockstate("minecraft:water")

        dirt_water = dirt + water

        self.assertNotEqual(dirt, dirt_water)
        self.assertIsNot(dirt, dirt_water)
        self.assertEqual(3, self.manager[dirt_water])

        with self.assertRaises(KeyError):
            random_block = self.manager[10000]

    def test_get_block_from_manager(self):
        dirt = Block.from_string_blockstate("minecraft:dirt")
        stone = Block.from_string_blockstate("minecraft:stone")
        granite = Block.from_string_blockstate("minecraft:granite")
        water = Block.from_string_blockstate("minecraft:water")
        dirt_water = dirt + water

        self.assertEqual(dirt, self.manager[0])
        self.assertEqual(stone, self.manager[1])
        self.assertEqual(granite, self.manager[2])
        self.assertEqual(dirt_water, self.manager[3])

        with self.assertRaises(KeyError):
            brain_coral = Block.from_string_blockstate("minecraft:brain_coral")
            internal_id = self.manager[brain_coral]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from typing import Tuple, Dict, Optional
import numpy

from amulet_nbt import (
    TAG_Compound,
    TAG_List,
    TAG_Byte,
    TAG_Int,
    TAG_Long_Array,
    TAG_String,
)

from amulet.api.chunk import Chunk
from amulet.api.registry import BiomeManager
from amulet.api.data_types import AnyNDArray, BiomeType
from amulet.utils.world_utils import (
    decode_long_array,
    encode_long_array,
)

from .anvil_2709 import (
    Anvil2709Interface as ParentInterface,
)


class Anvil2844Interface(ParentInterface):
    """
    Note that some of these changes happened in earlier snapshots
    Chunk restructuring
    Contents of Level tag moved into root
    Some tags renamed from PascalCase to snake_case
        2844
        Level.Entities -> entities.
        Level.TileEntities -> block_entities.
        Level.TileTicks and Level.ToBeTicked have moved to block_ticks.
        Level.LiquidTicks and Level.LiquidsToBeTicked have moved to fluid_ticks.
        Level.Sections -> sections.
        Level.Structures -> structures.
        Level.Structures.Starts -> structures.starts.
        Level.Sections[].block_states -> sections[].block_states.
        Level.Sections[].biomes -> sections[].biomes
        Added yPos the minimum section y position in the chunk.
        Added below_zero_retrogen containing data to support below zero generation.
        Added blending_data containing data to support blending new world generation with existing chunks.
        2836
        Level.Sections[].BlockStates & Level.Sections[].Palette -> Level.Sections[].block_states.
        Level.Biomes -> Level.Sections[].biomes.
        Level.CarvingMasks[] is now long[] instead of byte[].
    """

    BlockEntities = "block_entities"
    Entities = "entities"
    Sections = "sections"
    Structures = "structures"

    @staticmethod
    def minor_is_valid(key: int):
        return 2844 <= key < 2900

    def _decode_chunk(
        self, chunk: Chunk, root: TAG_Compound, bounds: Tuple[int, int]
    ) -> Tuple["Chunk", AnyNDArray]:
        self._decode_root(chunk, root)
        floor_cy = root.pop("yPos").value
        self._decode_level(chunk, root, bounds, floor_cy)
        sections = self._extract_sections(chunk, root)
        self._decode_sections(chunk, sections)
        palette = self._decode_blocks(chunk, sections)
        return chunk, palette

    def _decode_level(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int], floor_cy: int
    ):
        self._decode_coords(chunk, level)
        self._decode_last_update(chunk, level)
        self._decode_status(chunk, level)
        self._decode_inhabited_time(chunk, level)
        self._decode_height(chunk, level, bounds)
        self._decode_entities(chunk, level)
        self._decode_block_entities(chunk, level)
        self._decode_block_ticks(chunk, level, floor_cy)
        self._decode_fluid_ticks(chunk, level, floor_cy)
        self._decode_post_processing(chunk, level, floor_cy)
        self._decode_structures(chunk, level)

    def _decode_sections(self, chunk: Chunk, sections: Dict[int, TAG_Compound]):
        super()._decode_sections(chunk, sections)
        self._decode_biome_sections(chunk, sections)

    def _decode_block_section(
        self, section: TAG_Compound
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        block_states = self.get_obj(section, "block_states", TAG_Compound)
        if (
            isinstance(block_states, TAG_Compound) and "palette" in block_states
        ):  # 1.14 makes block_palette/blocks optional.
            section_palette = self._decode_block_palette(block_states.pop("palette"))
            data = block_states.pop("data", None)
            if data is None:
                arr = numpy.zeros((16, 16, 16), numpy.uint32)
            else:
                decoded = decode_long_array(
                    data.value,
                    16 ** 3,
                    max(4, (len(section_palette) - 1).bit_length()),
                    dense=self.LongArrayDense,
                ).astype(numpy.uint32)
                arr = numpy.transpose(decoded.reshape((16, 16, 16)), (2, 0, 1))
            return arr, section_palette
        else:
            return None

    @staticmethod
    def _decode_biome_palette(palette: TAG_List) -> list[BiomeType]:
        return [entry.value for entry in palette]

    def _decode_biome_section(
        self, section: TAG_Compound
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        biomes = self.get_obj(section, "biomes", TAG_Compound)
        if isinstance(biomes, TAG_Compound) and "palette" in biomes:
            section_palette = self._decode_biome_palette(biomes.pop("palette"))
            data = biomes.pop("data", None)
            if data is None:
                # TODO: in the new biome system just leave this as the number
                arr = numpy.zeros((4, 4, 4), numpy.uint32)
            else:
                arr = numpy.transpose(
                    decode_long_array(
                        data.value,
                        4 ** 3,
                        (len(section_palette) - 1).bit_length(),
                        dense=self.LongArrayDense,
                    )
                    .astype(numpy.uint32)
                    .reshape((4, 4, 4)),
                    (2, 0, 1),
                )
            return arr, section_palette
        else:
            return None

    def _decode_biome_sections(
        self, chunk: Chunk, chunk_sections: Dict[int, TAG_Compound]
    ):
        biomes: Dict[int, numpy.ndarray] = {}
        palette = BiomeManager()

        for cy, section in chunk_sections.items():
            data = self._decode_biome_section(section)
            if data is not None:
                arr, section_palette = data
                lut = numpy.array(
                    [palette.get_add_biome(biome) for biome in section_palette]
                )
                biomes[cy] = lut[arr].astype(numpy.uint32)

        chunk.biomes = biomes
        chunk.biome_palette = palette

    def _decode_block_ticks(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        chunk.misc.setdefault("block_ticks", {}).update(
            self._decode_ticks(self.get_obj(compound, "block_ticks", TAG_List))
        )

    def _decode_fluid_ticks(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        chunk.misc.setdefault("fluid_ticks", {}).update(
            self._decode_ticks(self.get_obj(compound, "fluid_ticks", TAG_List))
        )

    def _encode_chunk(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, int],
        bounds: Tuple[int, int],
    ) -> TAG_Compound:
        root = self._init_encode(chunk)
        self._encode_root(root, max_world_version)
        self._encode_level(chunk, root, bounds)
        sections = self._init_sections(chunk)
        self._encode_sections(chunk, sections, bounds)
        self._encode_blocks(chunk, sections, palette, bounds)
        sections_list = []
        for cy, section in sections.items():
            section["Y"] = TAG_Byte(cy)
            sections_list.append(section)
        root[self.Sections] = TAG_List(sections_list)
        return root

    def _encode_level(self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]):
        self._encode_coords(chunk, level, bounds)
        self._encode_last_update(chunk, level)
        self._encode_status(chunk, level)
        self._encode_inhabited_time(chunk, level)
        self._encode_height(chunk, level, bounds)
        self._encode_entities(chunk, level)
        self._encode_block_entities(chunk, level)
        self._encode_block_ticks(chunk, level, bounds)
        self._encode_fluid_ticks(chunk, level, bounds)
        self._encode_post_processing(chunk, level, bounds)
        self._encode_structures(chunk, level)

    def _encode_coords(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        super()._encode_coords(chunk, level, bounds)
        level["yPos"] = TAG_Int(bounds[0] >> 4)

    def _encode_sections(
        self, chunk: Chunk, sections: Dict[int, TAG_Compound], bounds: Tuple[int, int]
    ):
        super()._encode_sections(chunk, sections, bounds)
        self._encode_biome_sections(chunk, sections, bounds)

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        palette: AnyNDArray,
        cy: int,
    ):
        if cy in chunk.blocks:
            block_sub_array = numpy.transpose(
                chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
            ).ravel()

            sub_palette_, block_sub_array = numpy.unique(
                block_sub_array, return_inverse=True
            )
            sub_palette = self._encode_block_palette(palette[sub_palette_])
            section = sections.setdefault(cy, TAG_Compound())
            block_states = section["block_states"] = TAG_Compound(
                {"palette": sub_palette}
            )
            if len(sub_palette) != 1:
                block_states["data"] = TAG_Long_Array(
                    encode_long_array(
                        block_sub_array, dense=self.LongArrayDense, min_bits_per_entry=4
                    )
                )
            return True
        return False

    @staticmethod
    def _encode_biome_palette(palette: list[BiomeType]) -> TAG_List:
        return TAG_List([TAG_String(entry) for entry in palette])

    def _encode_biome_section(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        cy: int,
    ) -> bool:
        chunk.biomes.convert_to_3d()
        if cy in chunk.biomes:
            biome_sub_array = numpy.transpose(
                chunk.biomes.get_section(cy), (1, 2, 0)
            ).ravel()

            sub_palette_, biome_sub_array = numpy.unique(
                biome_sub_array, return_inverse=True
            )
            sub_palette = self._encode_biome_palette(chunk.biome_palette[sub_palette_])
            section = sections.setdefault(cy, TAG_Compound())
            biomes = section["biomes"] = TAG_Compound({"palette": sub_palette})
            if len(sub_palette) != 1:
                biomes["data"] = TAG_Long_Array(
                    encode_long_array(biome_sub_array, dense=self.LongArrayDense)
                )
            return True
        return False

    def _encode_biome_sections(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        bounds: Tuple[int, int],
    ):
        for cy in range(bounds[0] >> 4, bounds[1] >> 4):
            self._encode_biome_section(chunk, sections, cy)

    def _encode_block_ticks(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        compound["block_ticks"] = self._encode_ticks(chunk.misc.get("block_ticks", {}))

    def _encode_fluid_ticks(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        compound["fluid_ticks"] = self._encode_ticks(chunk.misc.get("fluid_ticks", {}))


export = Anvil2844Interface



from .leveldb_33 import (
    LevelDB33Interface as ParentInterface,
)


class LevelDB34Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 34)


export = LevelDB34Interface



import os.path
import unittest
import json

from tests.data.util import WorldTemp

from amulet import load_format
from tests.data import worlds_src
from tests.data.util import for_each_world, BaseWorldTest


@for_each_world(globals(), worlds_src.levels)
class TestWorldHeight(BaseWorldTest, unittest.TestCase):
    def test_height(self):
        with WorldTemp(self.WorldPath) as world_temp:
            test_data_path = os.path.join(world_temp.temp_path, "world_test_data.json")
            self.assertTrue(
                os.path.isfile(test_data_path),
                msg=f"Could not find world_test_data.json for world {self.WorldPath}",
            )
            with open(test_data_path) as f:
                test_data = json.load(f)
            self.assertIn("dim_height", test_data, msg=self.WorldPath)
            self.assertIsInstance(test_data["dim_height"], dict, msg=self.WorldPath)
            height_data = test_data["dim_height"]
            world_wrapper = load_format(world_temp.temp_path)
            world_wrapper.open()
            self.assertEqual(
                sorted(test_data["dim_height"]),
                sorted(world_wrapper.dimensions),
                msg=self.WorldPath,
            )
            for dim in world_wrapper.dimensions:
                box = world_wrapper.bounds(dim).to_box()
                self.assertEqual(
                    height_data[dim], [list(box.min), list(box.max)], msg=self.WorldPath
                )
            world_wrapper.close()


if __name__ == "__main__":
    unittest.main()



from .leveldb_25 import (
    LevelDB25Interface as ParentInterface,
)


class LevelDB26Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 26)


export = LevelDB26Interface



from .leveldb_27 import (
    LevelDB27Interface as ParentInterface,
)


class LevelDB28Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 28)


export = LevelDB28Interface



from __future__ import annotations

from typing import Dict, Set, Tuple, Optional, TYPE_CHECKING

import numpy
from amulet_nbt import (
    TAG_Compound,
    TAG_List,
    TAG_String,
    TAG_Long_Array,
    TAG_Long,
    TAG_Short,
)

import amulet
from amulet.api.data_types import AnyNDArray, BlockCoordinates
from amulet.api.block import Block
from amulet.api.chunk import StatusFormats
from .anvil_0 import Anvil0Interface
from amulet.utils.world_utils import (
    decode_long_array,
    encode_long_array,
)

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk


class Anvil1444Interface(Anvil0Interface):
    """
    Moved light and terrain populated to Status
    Made blocks paletted
    Added more tick tags
    Added structures tag
    """

    Structures = "Structures"
    LongArrayDense = True

    def __init__(self):
        super().__init__()
        self._set_feature("status", StatusFormats.Java_13)

    @staticmethod
    def minor_is_valid(key: int):
        return 1444 <= key < 1466

    def _decode_level(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int], floor_cy: int
    ):
        super()._decode_level(chunk, level, bounds, floor_cy)
        self._decode_fluid_ticks(chunk, level, floor_cy)
        self._decode_post_processing(chunk, level, floor_cy)
        self._decode_structures(chunk, level)

    def _decode_status(self, chunk: Chunk, compound: TAG_Compound):
        chunk.status = self.get_obj(
            compound, "Status", TAG_String, TAG_String("full")
        ).value

    def _decode_block_section(
        self, section: TAG_Compound
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        if "Palette" not in section:  # 1.14 makes block_palette/blocks optional.
            return None
        section_palette = self._decode_block_palette(section.pop("Palette"))
        decoded = decode_long_array(
            section.pop("BlockStates").value,
            4096,
            max(4, (len(section_palette) - 1).bit_length()),
            dense=self.LongArrayDense,
        ).astype(numpy.uint32)
        arr = numpy.transpose(decoded.reshape((16, 16, 16)), (2, 0, 1))
        return arr, section_palette

    def _decode_blocks(
        self, chunk: Chunk, chunk_sections: Dict[int, TAG_Compound]
    ) -> AnyNDArray:
        blocks: Dict[int, numpy.ndarray] = {}
        palette = [Block(namespace="minecraft", base_name="air")]

        for cy, section in chunk_sections.items():
            data = self._decode_block_section(section)
            if data is not None:
                arr, section_palette = data
                blocks[cy] = arr + len(palette)
                palette += section_palette

        np_palette, inverse = numpy.unique(palette, return_inverse=True)
        np_palette: numpy.ndarray
        inverse: numpy.ndarray
        inverse = inverse.astype(numpy.uint32)
        for cy in blocks:
            blocks[cy] = inverse[blocks[cy]]
        chunk.blocks = blocks
        return np_palette

    @staticmethod
    def _decode_block_palette(palette: TAG_List) -> list:
        blockstates = []
        for entry in palette:
            namespace, base_name = entry["Name"].value.split(":", 1)
            properties = entry.get("Properties", TAG_Compound({})).value
            block = Block(
                namespace=namespace, base_name=base_name, properties=properties
            )
            blockstates.append(block)
        return blockstates

    @staticmethod
    def _decode_to_be_ticked(ticks: TAG_List, floor_cy: int) -> Set[BlockCoordinates]:
        ticks_out = set()
        for cy, section_ticks in enumerate(ticks):
            block_cy = (floor_cy + cy) << 4
            for pos in section_ticks:
                # TODO: check if these are correct. The order may be wrong.
                x = pos & 0xF
                y = (pos >> 4) & 0xF
                z = (pos >> 8) & 0xF
                ticks_out.add((x, block_cy + y, z))
        return ticks_out

    def _decode_block_ticks(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        super()._decode_block_ticks(chunk, compound, floor_cy)
        chunk.misc["to_be_ticked"] = self._decode_to_be_ticked(
            self.get_obj(compound, "ToBeTicked", TAG_List), floor_cy
        )

    def _decode_fluid_ticks(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        chunk.misc.setdefault("fluid_ticks", {}).update(
            self._decode_ticks(self.get_obj(compound, "LiquidTicks", TAG_List))
        )
        chunk.misc["liquids_to_be_ticked"] = self._decode_to_be_ticked(
            self.get_obj(compound, "LiquidsToBeTicked", TAG_List), floor_cy
        )

    def _decode_post_processing(
        self, chunk: Chunk, compound: TAG_Compound, floor_cy: int
    ):
        chunk.misc["post_processing"] = self._decode_to_be_ticked(
            self.get_obj(compound, "PostProcessing", TAG_List), floor_cy
        )

    def _decode_structures(self, chunk: Chunk, compound: TAG_Compound):
        chunk.misc["structures"] = self.get_obj(compound, self.Structures, TAG_Compound)

    def _encode_level(self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]):
        super()._encode_level(chunk, level, bounds)
        self._encode_fluid_ticks(chunk, level, bounds)
        self._encode_post_processing(chunk, level, bounds)
        self._encode_structures(chunk, level)

    def _encode_status(self, chunk: Chunk, level: TAG_Compound):
        # Order the float value based on the order they would be run. Newer replacements for the same come just after
        # to save back find the next lowest valid value.
        status = chunk.status.as_type(self._features["status"])
        level["Status"] = TAG_String(status)

    def _encode_inhabited_time(self, chunk: Chunk, level: TAG_Compound):
        level["InhabitedTime"] = TAG_Long(chunk.misc.get("inhabited_time", 0))

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        palette: AnyNDArray,
        cy: int,
    ) -> bool:
        if cy in chunk.blocks:
            block_sub_array = numpy.transpose(
                chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
            ).ravel()

            sub_palette_, block_sub_array = numpy.unique(
                block_sub_array, return_inverse=True
            )
            sub_palette = self._encode_block_palette(palette[sub_palette_])
            if (
                len(sub_palette) == 1
                and sub_palette[0]["Name"].value == "minecraft:air"
            ):
                return False

            section = sections.setdefault(cy, TAG_Compound())
            section["BlockStates"] = TAG_Long_Array(
                encode_long_array(
                    block_sub_array, dense=self.LongArrayDense, min_bits_per_entry=4
                )
            )
            section["Palette"] = sub_palette
            return True
        return False

    def _encode_blocks(
        self,
        chunk: Chunk,
        sections: Dict[int, TAG_Compound],
        palette: AnyNDArray,
        bounds: Tuple[int, int],
    ):
        for cy in range(bounds[0] >> 4, bounds[1] >> 4):
            self._encode_block_section(chunk, sections, palette, cy)

    @staticmethod
    def _encode_block_palette(blockstates: list) -> TAG_List:
        palette = TAG_List()
        for block in blockstates:
            entry = TAG_Compound()
            entry["Name"] = TAG_String(f"{block.namespace}:{block.base_name}")
            if block.properties:
                entry["Properties"] = TAG_Compound(block.properties)
            palette.append(entry)
        return palette

    @staticmethod
    def _encode_to_be_ticked(
        ticks: Set[BlockCoordinates], bounds: Tuple[int, int]
    ) -> TAG_List:
        cy_min = bounds[0] >> 4
        cy_max = bounds[1] >> 4
        ticks_out = TAG_List([TAG_List([], 2) for _ in range(cy_min, cy_max)])
        if isinstance(ticks, set):
            for k in ticks:
                try:
                    (x, y, z) = k
                    cy = y >> 4
                    if cy_min <= cy < cy_max:
                        x = x & 15
                        y = y & 15
                        z = z & 15
                        ticks_out[cy].append(TAG_Short((z << 8) + (y << 4) + x))
                except Exception:
                    amulet.log.error(f"Could not serialise tick data {k}")
        return ticks_out

    def _encode_block_ticks(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        super()._encode_block_ticks(chunk, compound, bounds)
        compound["ToBeTicked"] = self._encode_to_be_ticked(
            chunk.misc.get("to_be_ticked"), bounds
        )

    def _encode_fluid_ticks(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        compound["LiquidTicks"] = self._encode_ticks(chunk.misc.get("fluid_ticks", {}))
        compound["LiquidsToBeTicked"] = self._encode_to_be_ticked(
            chunk.misc.get("liquids_to_be_ticked"), bounds
        )

    def _encode_post_processing(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        compound["PostProcessing"] = self._encode_to_be_ticked(
            chunk.misc.get("post_processing"), bounds
        )

    @staticmethod
    def _encode_structures(chunk: Chunk, compound: TAG_Compound):
        compound["Structures"] = chunk.misc.get(
            "structures",
            TAG_Compound(
                {
                    "References": TAG_Compound(),
                    "Starts": TAG_Compound(),
                }
            ),
        )


export = Anvil1444Interface



from .leveldb_22 import (
    LevelDB22Interface as ParentInterface,
)


class LevelDB23Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 23)


export = LevelDB23Interface



from typing import Any, List, Dict, Tuple, Optional, TYPE_CHECKING, Iterable, Union

from amulet.api.data_types import Dimension, PlatformType, ChunkCoordinates, AnyNDArray
from amulet.api.wrapper import FormatWrapper
from amulet.api.errors import ChunkDoesNotExist, PlayerDoesNotExist
from amulet.api.player import Player
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup
from amulet.api import wrapper as api_wrapper

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface


class VoidFormatWrapper(FormatWrapper):
    """
    A custom :class:`FormatWrapper` class that has no associated data.

    This is just to make the :class:`ImmutableStructure` class happy since it requires a :class:`FormatWrapper` class.

    All methods effectively do nothing.
    """

    @property
    def level_name(self) -> str:
        return "Void"

    @staticmethod
    def is_valid(path: str) -> bool:
        return False

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {}

    @property
    def dimensions(self) -> List[Dimension]:
        return ["main"]

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_identifier: Any):
        pass

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        raise Exception("If this is called something is wrong")

    def _encode(
        self,
        interface: api_wrapper.Interface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> Any:
        raise Exception("If this is called something is wrong")

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs
    ):
        pass

    def _open(self):
        pass

    def _save(self):
        pass

    def _close(self):
        pass

    def unload(self):
        pass

    def all_chunk_coords(self, dimension: Dimension) -> Iterable[ChunkCoordinates]:
        yield from ()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return False

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        pass

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        pass

    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        raise ChunkDoesNotExist

    def all_player_ids(self) -> Iterable[str]:
        yield from ()

    def has_player(self, player_id: str) -> bool:
        return False

    def _load_player(self, player_id: str) -> Player:
        raise PlayerDoesNotExist

    def _get_raw_player_data(self, player_id: str) -> Any:
        raise PlayerDoesNotExist



from __future__ import annotations

from .leveldb_9 import (
    LevelDB9Interface,
)


class LevelDB10Interface(LevelDB9Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 10)


export = LevelDB10Interface



from .leveldb_26 import (
    LevelDB26Interface as ParentInterface,
)


class LevelDB27Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 27)


export = LevelDB27Interface



from .leveldb_30 import (
    LevelDB30Interface as ParentInterface,
)


class LevelDB31Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 31)


export = LevelDB31Interface



# A cython (and python) wrapper for leveldb.

import os
import sys
from cpython.bytes cimport PyBytes_FromStringAndSize
from typing import Dict, Iterator, Tuple
cimport cython

if sys.platform == "win32":
    if sys.maxsize > 2 ** 32:  # 64 bit python
        lib_name = "leveldb_mcpe_win_amd64.dll"
    else:  # 32 bit python
        lib_name = "leveldb_mcpe_win32.dll"
elif sys.platform == "darwin":
    lib_name = "leveldb_mcpe_macosx_10_9_x86_64.dylib"
else:  # linux, compile your own .so if this errors!
    lib_name = "leveldb_mcpe_linux_x86_64.so"


lib_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), lib_name)
assert os.path.isfile(
    lib_path
), f"Could not find the leveldb shared library at {lib_path}"


IF UNAME_SYSNAME == "Windows":
    cdef extern from "<windows.h>":
        pass
    cdef extern from "<libloaderapi.h>":
        void* LoadLibraryA(char *lpLibFileName)
        void* GetProcAddress(void *hModule, char *lpProcName)

    cdef void *module = LoadLibraryA(lib_path.encode("utf-8"))

    cdef void *getFunction(char *functionName):
        return GetProcAddress(module, functionName)

ELIF UNAME_SYSNAME == "Darwin" or UNAME_SYSNAME == "Linux":
    cdef extern from "dlfcn.h":
        void *dlopen(char *filename, int mode)
        void *dlsym(void *handle, char *name)

        ctypedef enum DL_MODES:
            RTLD_LAZY
            RTLD_NOW
            RTLD_BINDING_MASK
            RTLD_NOLOAD
            RTLD_GLOBAL

    cdef void *module = dlopen(lib_path.encode("utf-8"), RTLD_LAZY)

    cdef void *getFunction(char *functionName):
        return dlsym(module, functionName)

ELSE:
    cdef void *getFunction(char *functionName):
        raise Exception(f"{os.uname()[0]} not yet supported")

cdef extern from "leveldb.h":
    struct leveldb_t:
        pass
    struct leveldb_cache_t:
        pass
    struct leveldb_comparator_t:
        pass
    struct leveldb_env_t:
        pass
    struct leveldb_filelock_t:
        pass
    struct leveldb_filterpolicy_t:
        pass
    struct leveldb_iterator_t:
        pass
    struct leveldb_logger_t:
        pass
    struct leveldb_options_t:
        pass
    struct leveldb_randomfile_t:
        pass
    struct leveldb_readoptions_t:
        pass
    struct leveldb_seqfile_t:
        pass
    struct leveldb_snapshot_t:
        pass
    struct leveldb_writablefile_t:
        pass
    struct leveldb_writebatch_t:
        pass
    struct leveldb_writeoptions_t:
        pass

    ctypedef enum compression:
        leveldb_no_compression=0,
        leveldb_zlib_compression=2,
        leveldb_zlib_raw_compression=4,


ctypedef leveldb_t* (*PFNleveldb_openPROC)(const leveldb_options_t* options, const char* name, char** errptr)
cdef PFNleveldb_openPROC leveldb_open = <PFNleveldb_openPROC>getFunction(b"leveldb_open")

ctypedef void (*PFNleveldb_closePROC)(leveldb_t* db)
cdef PFNleveldb_closePROC leveldb_close = <PFNleveldb_closePROC>getFunction(b"leveldb_close")

ctypedef void (*PFNleveldb_putPROC)(leveldb_t* db, const leveldb_writeoptions_t* options, const char* key, size_t keylen, const char* val, size_t vallen, char** errptr)
cdef PFNleveldb_putPROC leveldb_put = <PFNleveldb_putPROC>getFunction(b"leveldb_put")

ctypedef void (*PFNleveldb_deletePROC)(leveldb_t* db, const leveldb_writeoptions_t* options, const char* key, size_t keylen, char** errptr)
cdef PFNleveldb_deletePROC leveldb_delete = <PFNleveldb_deletePROC>getFunction(b"leveldb_delete")

ctypedef void (*PFNleveldb_writePROC)(leveldb_t* db, const leveldb_writeoptions_t* options, leveldb_writebatch_t* batch, char** errptr)
cdef PFNleveldb_writePROC leveldb_write = <PFNleveldb_writePROC>getFunction(b"leveldb_write")

ctypedef char* (*PFNleveldb_getPROC)(leveldb_t* db, const leveldb_readoptions_t* options, const char* key, size_t keylen, size_t* vallen, char** errptr)
cdef PFNleveldb_getPROC leveldb_get = <PFNleveldb_getPROC>getFunction(b"leveldb_get")

ctypedef leveldb_iterator_t* (*PFNleveldb_create_iteratorPROC)(leveldb_t* db, const leveldb_readoptions_t* options)
cdef PFNleveldb_create_iteratorPROC leveldb_create_iterator = <PFNleveldb_create_iteratorPROC>getFunction(b"leveldb_create_iterator")

ctypedef const leveldb_snapshot_t* (*PFNleveldb_create_snapshotPROC)(leveldb_t* db)
cdef PFNleveldb_create_snapshotPROC leveldb_create_snapshot = <PFNleveldb_create_snapshotPROC>getFunction(b"leveldb_create_snapshot")

ctypedef void (*PFNleveldb_release_snapshotPROC)(leveldb_t* db, const leveldb_snapshot_t* snapshot)
cdef PFNleveldb_release_snapshotPROC leveldb_release_snapshot = <PFNleveldb_release_snapshotPROC>getFunction(b"leveldb_release_snapshot")

ctypedef char* (*PFNleveldb_property_valuePROC)(leveldb_t* db, const char* propname)

ctypedef void (*PFNleveldb_approximate_sizesPROC)(leveldb_t* db, int num_ranges, const char* range_start_key, const size_t* range_start_key_len, const char* range_limit_key, const size_t* range_limit_key_len, long long* sizes)
cdef PFNleveldb_approximate_sizesPROC leveldb_approximate_sizes = <PFNleveldb_approximate_sizesPROC>getFunction(b"leveldb_approximate_sizes")

ctypedef void (*PFNleveldb_compact_rangePROC)(leveldb_t* db, const char* start_key, size_t start_key_len, const char* limit_key, size_t limit_key_len)
cdef PFNleveldb_compact_rangePROC leveldb_compact_range = <PFNleveldb_compact_rangePROC>getFunction(b"leveldb_compact_range")

ctypedef void (*PFNleveldb_destroy_dbPROC)(const leveldb_options_t* options, const char* name, char** errptr)

ctypedef void (*PFNleveldb_repair_dbPROC)(const leveldb_options_t* options, const char* name, char** errptr)

ctypedef void (*PFNleveldb_iter_destroyPROC)(leveldb_iterator_t* iter)
cdef PFNleveldb_iter_destroyPROC leveldb_iter_destroy = <PFNleveldb_iter_destroyPROC>getFunction(b"leveldb_iter_destroy")

ctypedef unsigned char (*PFNleveldb_iter_validPROC)(const leveldb_iterator_t* iter)
cdef PFNleveldb_iter_validPROC leveldb_iter_valid = <PFNleveldb_iter_validPROC>getFunction(b"leveldb_iter_valid")

ctypedef void (*PFNleveldb_iter_seek_to_firstPROC)(leveldb_iterator_t* iter)
cdef PFNleveldb_iter_seek_to_firstPROC leveldb_iter_seek_to_first = <PFNleveldb_iter_seek_to_firstPROC>getFunction(b"leveldb_iter_seek_to_first")

ctypedef void (*PFNleveldb_iter_seek_to_lastPROC)(leveldb_iterator_t* iter)
cdef PFNleveldb_iter_seek_to_lastPROC leveldb_iter_seek_to_last = <PFNleveldb_iter_seek_to_lastPROC>getFunction(b"leveldb_iter_seek_to_last")

ctypedef void (*PFNleveldb_iter_seekPROC)(leveldb_iterator_t* iter, const char* k, size_t klen)
cdef PFNleveldb_iter_seekPROC leveldb_iter_seek = <PFNleveldb_iter_seekPROC>getFunction(b"leveldb_iter_seek")

ctypedef void (*PFNleveldb_iter_nextPROC)(leveldb_iterator_t* iter)
cdef PFNleveldb_iter_nextPROC leveldb_iter_next = <PFNleveldb_iter_nextPROC>getFunction(b"leveldb_iter_next")

ctypedef void (*PFNleveldb_iter_prevPROC)(leveldb_iterator_t* iter)
cdef PFNleveldb_iter_prevPROC leveldb_iter_prev = <PFNleveldb_iter_prevPROC>getFunction(b"leveldb_iter_prev")

ctypedef const char* (*PFNleveldb_iter_keyPROC)(const leveldb_iterator_t* iter, size_t* klen)
cdef PFNleveldb_iter_keyPROC leveldb_iter_key = <PFNleveldb_iter_keyPROC>getFunction(b"leveldb_iter_key")

ctypedef const char* (*PFNleveldb_iter_valuePROC)(const leveldb_iterator_t* iter, size_t* vlen)
cdef PFNleveldb_iter_valuePROC leveldb_iter_value = <PFNleveldb_iter_valuePROC>getFunction(b"leveldb_iter_value")

ctypedef void (*PFNleveldb_iter_get_errorPROC)(const leveldb_iterator_t* iter, char** errptr)
cdef PFNleveldb_iter_get_errorPROC leveldb_iter_get_error = <PFNleveldb_iter_get_errorPROC>getFunction(b"leveldb_iter_get_error")

ctypedef leveldb_writebatch_t* (*PFNleveldb_writebatch_createPROC)()
cdef PFNleveldb_writebatch_createPROC leveldb_writebatch_create = <PFNleveldb_writebatch_createPROC>getFunction(b"leveldb_writebatch_create")

ctypedef void (*PFNleveldb_writebatch_destroyPROC)(leveldb_writebatch_t* b)
cdef PFNleveldb_writebatch_destroyPROC leveldb_writebatch_destroy = <PFNleveldb_writebatch_destroyPROC>getFunction(b"leveldb_writebatch_destroy")

ctypedef void (*PFNleveldb_writebatch_clearPROC)(leveldb_writebatch_t* b)
cdef PFNleveldb_writebatch_clearPROC leveldb_writebatch_clear = <PFNleveldb_writebatch_clearPROC>getFunction(b"leveldb_writebatch_clear")

ctypedef void (*PFNleveldb_writebatch_putPROC)(leveldb_writebatch_t* b, const char* key, size_t klen, const char* val, size_t vlen)
cdef PFNleveldb_writebatch_putPROC leveldb_writebatch_put = <PFNleveldb_writebatch_putPROC>getFunction(b"leveldb_writebatch_put")

ctypedef void (*PFNleveldb_writebatch_deletePROC)(leveldb_writebatch_t* b, const char* key, size_t klen)
cdef PFNleveldb_writebatch_deletePROC leveldb_writebatch_delete = <PFNleveldb_writebatch_deletePROC>getFunction(b"leveldb_writebatch_delete")

ctypedef void (*PFNleveldb_writebatch_iteratePROC)(leveldb_writebatch_t* b, void* state, void (*put)(void*, const char* k, size_t klen, const char* v, size_t vlen), void (*deleted)(void*, const char* k, size_t klen))

ctypedef leveldb_options_t* (*PFNleveldb_options_createPROC)()
cdef PFNleveldb_options_createPROC leveldb_options_create = <PFNleveldb_options_createPROC>getFunction(b"leveldb_options_create")

ctypedef void (*PFNleveldb_options_destroyPROC)(leveldb_options_t* options)
cdef PFNleveldb_options_destroyPROC leveldb_options_destroy = <PFNleveldb_options_destroyPROC>getFunction(b"leveldb_options_destroy")

ctypedef void (*PFNleveldb_options_set_comparatorPROC)(leveldb_options_t* options, leveldb_comparator_t* cmp)

ctypedef void (*PFNleveldb_options_set_filter_policyPROC)(leveldb_options_t* options, leveldb_filterpolicy_t* policy)
cdef PFNleveldb_options_set_filter_policyPROC leveldb_options_set_filter_policy = <PFNleveldb_options_set_filter_policyPROC>getFunction(b"leveldb_options_set_filter_policy")

ctypedef void (*PFNleveldb_options_set_create_if_missingPROC)(leveldb_options_t* options, unsigned char v)
cdef PFNleveldb_options_set_create_if_missingPROC leveldb_options_set_create_if_missing = <PFNleveldb_options_set_create_if_missingPROC>getFunction(b"leveldb_options_set_create_if_missing")

ctypedef void (*PFNleveldb_options_set_error_if_existsPROC)(leveldb_options_t* options, unsigned char v)
cdef PFNleveldb_options_set_error_if_existsPROC leveldb_options_set_error_if_exists = <PFNleveldb_options_set_error_if_existsPROC>getFunction(b"leveldb_options_set_error_if_exists")

ctypedef void (*PFNleveldb_options_set_paranoid_checksPROC)(leveldb_options_t* options, unsigned char v)
cdef PFNleveldb_options_set_paranoid_checksPROC leveldb_options_set_paranoid_checks = <PFNleveldb_options_set_paranoid_checksPROC>getFunction(b"leveldb_options_set_paranoid_checks")

ctypedef void (*PFNleveldb_options_set_envPROC)(leveldb_options_t* options, leveldb_env_t* env)

ctypedef void (*PFNleveldb_options_set_info_logPROC)(leveldb_options_t* options, leveldb_logger_t* l)

ctypedef void (*PFNleveldb_options_set_write_buffer_sizePROC)(leveldb_options_t* options, size_t s)
cdef PFNleveldb_options_set_write_buffer_sizePROC leveldb_options_set_write_buffer_size = <PFNleveldb_options_set_write_buffer_sizePROC>getFunction(b"leveldb_options_set_write_buffer_size")

ctypedef void (*PFNleveldb_options_set_max_open_filesPROC)(leveldb_options_t* options, int n)
cdef PFNleveldb_options_set_max_open_filesPROC leveldb_options_set_max_open_files = <PFNleveldb_options_set_max_open_filesPROC>getFunction(b"leveldb_options_set_max_open_files")

ctypedef void (*PFNleveldb_options_set_cachePROC)(leveldb_options_t* options, leveldb_cache_t* c)
cdef PFNleveldb_options_set_cachePROC leveldb_options_set_cache = <PFNleveldb_options_set_cachePROC>getFunction(b"leveldb_options_set_cache")

ctypedef void (*PFNleveldb_options_set_block_sizePROC)(leveldb_options_t* options, size_t s)
cdef PFNleveldb_options_set_block_sizePROC leveldb_options_set_block_size = <PFNleveldb_options_set_block_sizePROC>getFunction(b"leveldb_options_set_block_size")

ctypedef void (*PFNleveldb_options_set_block_restart_intervalPROC)(leveldb_options_t* options, int n)

ctypedef void (*PFNleveldb_options_set_compressionPROC)(leveldb_options_t* options, int t)
cdef PFNleveldb_options_set_compressionPROC leveldb_options_set_compression = <PFNleveldb_options_set_compressionPROC>getFunction(b"leveldb_options_set_compression")

ctypedef leveldb_comparator_t* (*PFNleveldb_comparator_createPROC)(void* state, void (*destructor)(void*), int (*compare)(void*, const char* a, size_t alen, const char* b, size_t blen), const char* (*name)(void*))

ctypedef void (*PFNleveldb_comparator_destroyPROC)(leveldb_comparator_t* cmp)

# ctypedef leveldb_filterpolicy_t* (*PFNleveldb_filterpolicy_createPROC)(void* state, void (*destructor)(void*), char* (*create_filter)(void*, const char* const* key_array, const size_t* key_length_array, int num_keys, size_t* filter_length), unsigned char (*key_may_match)(void*, const char* key, size_t length, const char* filter, size_t filter_length), const char* (*name)(void*))

ctypedef void (*PFNleveldb_filterpolicy_destroyPROC)(leveldb_filterpolicy_t* policy)
cdef PFNleveldb_filterpolicy_destroyPROC leveldb_filterpolicy_destroy = <PFNleveldb_filterpolicy_destroyPROC>getFunction(b"leveldb_filterpolicy_destroy")

ctypedef leveldb_filterpolicy_t* (*PFNleveldb_filterpolicy_create_bloomPROC)(int bits_per_key)
cdef PFNleveldb_filterpolicy_create_bloomPROC leveldb_filterpolicy_create_bloom = <PFNleveldb_filterpolicy_create_bloomPROC>getFunction(b"leveldb_filterpolicy_create_bloom")

ctypedef leveldb_readoptions_t* (*PFNleveldb_readoptions_createPROC)()
cdef PFNleveldb_readoptions_createPROC leveldb_readoptions_create = <PFNleveldb_readoptions_createPROC>getFunction(b"leveldb_readoptions_create")

ctypedef void (*PFNleveldb_readoptions_destroyPROC)(leveldb_readoptions_t* options)
cdef PFNleveldb_readoptions_destroyPROC leveldb_readoptions_destroy = <PFNleveldb_readoptions_destroyPROC>getFunction(b"leveldb_readoptions_destroy")

ctypedef void (*PFNleveldb_readoptions_set_verify_checksumsPROC)(leveldb_readoptions_t* options, unsigned char v)
cdef PFNleveldb_readoptions_set_verify_checksumsPROC leveldb_readoptions_set_verify_checksums = <PFNleveldb_readoptions_set_verify_checksumsPROC>getFunction(b"leveldb_readoptions_set_verify_checksums")

ctypedef void (*PFNleveldb_readoptions_set_fill_cachePROC)(leveldb_readoptions_t* options, unsigned char v)
cdef PFNleveldb_readoptions_set_fill_cachePROC leveldb_readoptions_set_fill_cache = <PFNleveldb_readoptions_set_fill_cachePROC>getFunction(b"leveldb_readoptions_set_fill_cache")

ctypedef void (*PFNleveldb_readoptions_set_snapshotPROC)(leveldb_readoptions_t* options, const leveldb_snapshot_t*)
cdef PFNleveldb_readoptions_set_snapshotPROC leveldb_readoptions_set_snapshot = <PFNleveldb_readoptions_set_snapshotPROC>getFunction(b"leveldb_readoptions_set_snapshot")

ctypedef leveldb_writeoptions_t* (*PFNleveldb_writeoptions_createPROC)()
cdef PFNleveldb_writeoptions_createPROC leveldb_writeoptions_create = <PFNleveldb_writeoptions_createPROC>getFunction(b"leveldb_writeoptions_create")

ctypedef void (*PFNleveldb_writeoptions_destroyPROC)(leveldb_writeoptions_t* options)
cdef PFNleveldb_writeoptions_destroyPROC leveldb_writeoptions_destroy = <PFNleveldb_writeoptions_destroyPROC>getFunction(b"leveldb_writeoptions_destroy")

ctypedef void (*PFNleveldb_writeoptions_set_syncPROC)(leveldb_writeoptions_t* options, unsigned char v)
cdef PFNleveldb_writeoptions_set_syncPROC leveldb_writeoptions_set_sync = <PFNleveldb_writeoptions_set_syncPROC>getFunction(b"leveldb_writeoptions_set_sync")

ctypedef leveldb_cache_t* (*PFNleveldb_cache_create_lruPROC)(size_t capacity)
cdef PFNleveldb_cache_create_lruPROC leveldb_cache_create_lru = <PFNleveldb_cache_create_lruPROC>getFunction(b"leveldb_cache_create_lru")

ctypedef void (*PFNleveldb_cache_destroyPROC)(leveldb_cache_t* cache)
cdef PFNleveldb_cache_destroyPROC leveldb_cache_destroy = <PFNleveldb_cache_destroyPROC>getFunction(b"leveldb_cache_destroy")

ctypedef leveldb_env_t* (*PFNleveldb_create_default_envPROC)()

ctypedef void (*PFNleveldb_env_destroyPROC)(leveldb_env_t* env)

ctypedef void (*PFNleveldb_freePROC)(void* ptr)
cdef PFNleveldb_freePROC leveldb_free = <PFNleveldb_freePROC>getFunction(b"leveldb_free")

ctypedef int (*PFNleveldb_major_versionPROC)()

ctypedef int (*PFNleveldb_minor_versionPROC)()


class LevelDBException(Exception):
    """
    The exception thrown for all leveldb related errors.
    """

    pass


cdef inline void _checkError(char *err):
    """Utility function for checking the error code returned by some leveldb functions."""
    cdef bytes message
    if err is not NULL:  # Not an empty null-terminated string
        message = err
        leveldb_free(err)
        raise LevelDBException(message.decode("utf-8"))

cdef inline void _check_db(void *db):
    if db is NULL:
        raise LevelDBException("The database has been closed.")


@cython.final
cdef class LevelDB:
    cdef leveldb_t* db
    cdef leveldb_readoptions_t* read_options
    cdef leveldb_writeoptions_t* write_options

    def __init__(self, str path, unsigned char create_if_missing = False):
        """
        Construct a new :class:`LevelDB` instance from the database at the given path.

        A leveldb database is like a dictionary that only contains bytes as the keys and values and exists entirely on the disk.

        :param path: The path to the database directory.
        :param create_if_missing: If True and there is no database at the given path a new database will be created.
        :raises: LevelDBException if create_if_missing is False and the db does not exist.
        """
        self._open(path, create_if_missing)

    def __cinit__(self):
        self.db = NULL
        self.read_options = NULL
        self.write_options = NULL

    cdef void _open(self, str path, unsigned char create_if_missing = False) except *:
        if not os.path.isdir(path):
            if create_if_missing:
                os.makedirs(path)
            else:
                raise LevelDBException(f"No database exists to open at {path}")
        filter_policy = leveldb_filterpolicy_create_bloom(10)
        cache = leveldb_cache_create_lru(40 * 1024 * 1024)
        options = leveldb_options_create()
        # Many of these options were pulled from Podshot/MCEdit-Unified
        leveldb_options_set_compression(options, leveldb_zlib_raw_compression)
        leveldb_options_set_filter_policy(options, filter_policy)
        leveldb_options_set_create_if_missing(options, create_if_missing)
        leveldb_options_set_write_buffer_size(options, 4 * 1024 * 1024)
        leveldb_options_set_cache(options, cache)
        leveldb_options_set_block_size(options, 163840)
        cdef char *error = NULL
        db = leveldb_open(options, path.encode("utf-8"), &error)
        leveldb_options_destroy(options)
        _checkError(error)

        self.db = db
        self.read_options = leveldb_readoptions_create()
        self.write_options = leveldb_writeoptions_create()

    cpdef void close(self, unsigned char compact=True) except *:
        """
        Close the leveldb database.

        :param compact: If True will compact the database making it take less memory.
        """
        _check_db(self.db)
        if compact:
            leveldb_compact_range(self.db, NULL, 0, NULL, 0)
        leveldb_close(self.db)
        self.db = NULL
        leveldb_readoptions_destroy(self.read_options)
        leveldb_writeoptions_destroy(self.write_options)

    cpdef bytes get(self, bytes key):
        """
        Get a key from the database.

        :param key: The key to get from the database.
        :return: The data stored behind the given key.
        :raises: KeyError if the requested key is not present.
        """
        cdef size_t vallen = 0
        cdef char *val = self.get_raw(key, len(key), &vallen)
        ret = PyBytes_FromStringAndSize(val, vallen)
        leveldb_free(val)
        return ret

    cdef char* get_raw(self, char *key, size_t keylen, size_t *vallen) except *:
        """
        Low level cython wrapper for the get method.
        Return value must be freed using leveldb_free.
        
        :param key: The key to get from the database.
        :param keylen: The length of the key.
        :param vallen: size_t pointer to put the length of the returned val
        :return: The data stored behind the given key.
        :raises: KeyError if the requested key is not present.
        """
        _check_db(self.db)
        cdef char *error = NULL
        cdef char *val = leveldb_get(self.db, self.read_options, key, keylen, vallen, &error)
        _checkError(error)
        if val is NULL:
            raise KeyError(key)
        return val

    cpdef void put(self, bytes key, bytes val) except *:
        """
        Add a key to the database.

        :param key: The key to store the value under.
        :param val: The value to store.
        """
        self.put_raw(key, len(key), val, len(val))

    cdef void put_raw(self, char *key, size_t keylen, char *val, size_t vallen) except *:
        _check_db(self.db)
        cdef char *error = NULL
        leveldb_put(self.db, self.write_options, key, keylen, val, vallen, &error)
        _checkError(error)

    cpdef void putBatch(self, dict data: Dict[bytes, bytes]):
        """
        Put one or more key and value pair into the database. Works the same as dict.update
        
        :param data: A dictionary of keys and values to add to the database
        """

        _check_db(self.db)
        cdef bytes k, v
        batch = leveldb_writebatch_create()
        for k, v in data.items():
            leveldb_writebatch_put(batch, k, len(k), v, len(v))
        cdef char *error = NULL
        leveldb_write(self.db, self.write_options, batch, &error)
        _checkError(error)

    cpdef void delete(self, bytes key) except *:
        """
        Delete a key from the database.

        :param key: The key to delete from the database.
        """
        self.delete_raw(key, len(key))

    cdef void delete_raw(self, char *key, size_t keylen) except *:
        cdef char *error = NULL
        leveldb_delete(self.db, self.write_options, key, keylen, &error)
        _checkError(error)

    def iterate(
        self, start: bytes = None, end: bytes = None
    ):# -> Iterator[Tuple[bytes, bytes]]:
        """
        Iterate through all keys and data that exist between the given keys.

        :param start: The key to start at. Leave as None to start at the beginning.
        :param end: The key to end at. Leave as None to start at the beginning.
        """
        _check_db(self.db)
        cdef bytes key_bytes
        cdef char *key
        cdef char *val
        cdef size_t keylen, vallen
        it = leveldb_create_iterator(self.db, self.read_options)
        if start is None:
            leveldb_iter_seek_to_first(it)
        else:
            leveldb_iter_seek(it, start, len(start))
        try:
            while leveldb_iter_valid(it):
                key = leveldb_iter_key(it, &keylen)
                key_bytes = PyBytes_FromStringAndSize(key, keylen)
                if end is not None and key_bytes >= end:
                    break
                val = leveldb_iter_value(it, &vallen)
                yield key_bytes, PyBytes_FromStringAndSize(val, vallen)
                leveldb_iter_next(it)
        finally:
            leveldb_iter_destroy(it)

    def keys(self):# -> Iterator[bytes]:
        """An iterable of all the keys in the database."""
        _check_db(self.db)
        cdef char *key
        cdef size_t keylen
        it = leveldb_create_iterator(self.db, self.read_options)
        leveldb_iter_seek_to_first(it)
        try:
            while leveldb_iter_valid(it):
                key = leveldb_iter_key(it, &keylen)
                yield PyBytes_FromStringAndSize(key, keylen)
                leveldb_iter_next(it)
        finally:
            leveldb_iter_destroy(it)

    def items(self):# -> Iterator[Tuple[bytes, bytes]]:
        _check_db(self.db)
        cdef char *key
        cdef char *val
        cdef size_t keylen, vallen
        it = leveldb_create_iterator(self.db, self.read_options)
        leveldb_iter_seek_to_first(it)
        try:
            while leveldb_iter_valid(it):
                key = leveldb_iter_key(it, &keylen)
                val = leveldb_iter_value(it, &vallen)
                yield PyBytes_FromStringAndSize(key, keylen), PyBytes_FromStringAndSize(val, vallen)
                leveldb_iter_next(it)
        finally:
            leveldb_iter_destroy(it)

    def __contains__(self, bytes key):
        keys = list(self.iterate(key, key + b"\x00"))
        return bool(keys) and keys[0][0] == key

    def __iter__(self) -> Iterator[bytes]:
        return self.keys()



from .leveldb_29 import (
    LevelDB29Interface as ParentInterface,
)


class LevelDB30Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 30)


export = LevelDB30Interface



from .leveldb_37 import (
    LevelDB37Interface as ParentInterface,
)


class LevelDB38Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 38)


export = LevelDB38Interface



from __future__ import annotations

from typing import Tuple
import numpy
from amulet_nbt import TAG_Int_Array, TAG_Compound
from amulet.api.chunk import Chunk
from .anvil_1466 import (
    Anvil1466Interface,
)


class Anvil1467Interface(Anvil1466Interface):
    """
    Biomes now stored in an int array
    """

    @staticmethod
    def minor_is_valid(key: int):
        return 1467 <= key < 1484

    def _decode_biomes(self, chunk: Chunk, compound: TAG_Compound, floor_cy: int):
        biomes = compound.pop("Biomes", None)
        if isinstance(biomes, TAG_Int_Array) and biomes.value.size == 256:
            chunk.biomes = biomes.astype(numpy.uint32).reshape((16, 16))

    def _encode_biomes(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        if chunk.status.value > -0.7:
            chunk.biomes.convert_to_2d()
            level["Biomes"] = TAG_Int_Array(
                chunk.biomes.astype(dtype=numpy.uint32).ravel()
            )


export = Anvil1467Interface



import unittest

from tests.data.util import WorldTemp

from amulet import load_level, load_format
from amulet.level.formats.anvil_world import AnvilFormat
from amulet.api.level import World
from tests.data import worlds_src


class DefinitionBasedLoaderTestCase(unittest.TestCase):
    def test_identify(self):
        for path in worlds_src.JavaVanillaLevels:
            with WorldTemp(path) as world_temp:
                java_format = load_format(world_temp.temp_path)
                self.assertIsInstance(java_format, AnvilFormat)

    def test_loading(self):
        for path in worlds_src.levels:
            with WorldTemp(path) as world_temp:
                numerical_java_world = load_level(world_temp.temp_path)
                self.assertIsInstance(numerical_java_world, World)
                numerical_java_world.close()


if __name__ == "__main__":
    unittest.main()



from .leveldb_35 import (
    LevelDB35Interface as ParentInterface,
)


class LevelDB36Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 36)


export = LevelDB36Interface



from __future__ import annotations

from .leveldb_2 import (
    LevelDB2Interface,
)


class LevelDB3Interface(LevelDB2Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 3)
        self._set_feature("terrain", "2farray")


export = LevelDB3Interface



import unittest
from amulet.api.selection import SelectionGroup, SelectionBox


class SelectionTestCase(unittest.TestCase):
    def test_equals(self):
        box_1 = SelectionBox((0, 0, 0), (5, 5, 5))
        box_2 = SelectionBox((5, 5, 5), (0, 0, 0))
        self.assertEqual(box_1, box_2)

    def test_intersects(self):
        box_1 = SelectionBox((0, 0, 0), (5, 5, 5))
        box_2 = SelectionBox((4, 4, 4), (10, 10, 10))
        box_3 = SelectionBox((5, 5, 5), (10, 10, 10))
        box_4 = SelectionBox((1, 20, 1), (4, 25, 4))

        self.assertTrue(box_1.intersects(box_2))
        self.assertTrue(box_2.intersects(box_1))
        self.assertFalse(box_1.intersects(box_3))
        self.assertFalse(box_3.intersects(box_1))
        self.assertFalse(box_1.intersects(box_4))
        self.assertFalse(box_4.intersects(box_1))

    def test_is_contiguous(self):
        box = SelectionBox((0, 0, 0), (5, 5, 5))

        self.assertTrue(SelectionGroup(box).is_contiguous)
        # top corner
        self.assertTrue(
            SelectionGroup((box, SelectionBox((5, 5, 5), (10, 10, 10)))).is_contiguous
        )
        # bottom corner
        self.assertTrue(
            SelectionGroup((box, SelectionBox((-5, -5, -5), (0, 0, 0)))).is_contiguous
        )
        # top face
        self.assertTrue(
            SelectionGroup((box, SelectionBox((0, 5, 0), (5, 10, 5)))).is_contiguous
        )
        # bottom face
        self.assertTrue(
            SelectionGroup((box, SelectionBox((0, -5, 0), (5, 0, 5)))).is_contiguous
        )
        # edge
        self.assertTrue(
            SelectionGroup((box, SelectionBox((0, 5, 5), (5, 10, 10)))).is_contiguous
        )
        # edge partial
        self.assertTrue(
            SelectionGroup((box, SelectionBox((1, 5, 5), (4, 10, 10)))).is_contiguous
        )

        # disconnected top corner
        self.assertFalse(
            SelectionGroup((box, SelectionBox((6, 6, 6), (10, 10, 10)))).is_contiguous
        )
        # intersecting top corner
        self.assertFalse(
            SelectionGroup((box, SelectionBox((4, 4, 4), (10, 10, 10)))).is_contiguous
        )
        # intersecting top face
        self.assertFalse(
            SelectionGroup((box, SelectionBox((0, 4, 0), (5, 10, 5)))).is_contiguous
        )
        # intersecting bottom face
        self.assertFalse(
            SelectionGroup((box, SelectionBox((0, -4, 0), (5, 1, 5)))).is_contiguous
        )
        # intersecting on only two faces
        self.assertFalse(
            SelectionGroup((box, SelectionBox((0, 5, 6), (5, 10, 10)))).is_contiguous
        )

    def test_is_rectangular(self):
        box_1 = SelectionBox((0, 0, 0), (5, 5, 5))
        box_2 = SelectionBox((0, 5, 0), (5, 10, 5))
        box_3 = SelectionBox((0, 0, 5), (5, 5, 10))
        box_4 = SelectionBox((0, 5, 5), (5, 10, 10))

        self.assertTrue(SelectionGroup(box_1).is_rectangular)
        self.assertTrue(SelectionGroup((box_1, box_2)).is_rectangular)
        self.assertFalse(SelectionGroup((box_1, box_2, box_3)).is_rectangular)
        self.assertTrue(SelectionGroup((box_1, box_2, box_3, box_4)).is_rectangular)
        self.assertTrue(
            SelectionGroup((box_1, box_2, box_3, box_4, box_2)).is_rectangular
        )

    def test_single_block_box(self):
        box_1 = SelectionBox((0, 0, 0), (1, 1, 2))

        self.assertEqual((1, 1, 2), box_1.shape)
        self.assertEqual(2, len([x for x in box_1]))

        self.assertIn((0, 0, 0), box_1)
        self.assertNotIn((1, 1, 2), box_1)

    def test_sorted_iterator(self):
        box_1 = SelectionBox((0, 0, 0), (4, 4, 4))
        box_2 = SelectionBox((7, 7, 7), (10, 10, 10))
        box_3 = SelectionBox((15, 15, 15), (20, 20, 20))

        selection_1 = SelectionGroup(
            (
                box_1,
                box_2,
                box_3,
            )
        )
        selection_2 = SelectionGroup(
            (
                box_1,
                box_3,
                box_2,
            )
        )

        self.assertEqual(selection_1, selection_2)

    def test_subtract(self):
        box_1 = SelectionGroup(
            SelectionBox(
                (0, 0, 0),
                (32, 32, 32),
            )
        )
        box_2 = SelectionGroup(
            SelectionBox(
                (0, 0, 0),
                (16, 16, 16),
            )
        )
        box_3 = box_1.subtract(box_2)
        box_4 = SelectionGroup(
            (
                SelectionBox((0, 16, 0), (32, 32, 32)),
                SelectionBox((0, 0, 16), (32, 16, 32)),
                SelectionBox((16, 0, 0), (32, 16, 16)),
            )
        )
        self.assertEqual(box_3, box_4)


if __name__ == "__main__":
    unittest.main()



from .leveldb_23 import (
    LevelDB23Interface as ParentInterface,
)


class LevelDB24Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 24)


export = LevelDB24Interface



from __future__ import annotations

from .leveldb_38 import (
    LevelDB38Interface as ParentInterface,
)


class LevelDB39Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 39)
        self._set_feature("data_2d", "height512|biome4096")


export = LevelDB39Interface



from __future__ import annotations

from .leveldb_7 import (
    LevelDB7Interface,
)


class LevelDB8Interface(LevelDB7Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 8)
        self._set_feature("terrain", "2fnpalette")


export = LevelDB8Interface



from __future__ import annotations

from .leveldb_24 import (
    LevelDB24Interface as ParentInterface,
)


class LevelDB25Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 25)


export = LevelDB25Interface



import unittest
import json
import numpy

from amulet.utils.world_utils import decode_long_array, encode_long_array
from tests.data.util import get_data_path


class LongArrayTestCase(unittest.TestCase):
    def test_longarray(self):
        with open(get_data_path("longarraytest.json")) as json_data:
            test_data = json.load(json_data)

        test_ran = False
        for test_entry in test_data["tests"]:
            test_ran = True
            block_array = numpy.asarray(test_entry["block_array"])
            long_array = numpy.asarray(test_entry["long_array"])
            palette_size = test_entry["palette_size"]

            numpy.testing.assert_array_equal(
                block_array,
                decode_long_array(
                    long_array, len(block_array), (palette_size - 1).bit_length()
                ),
            )

            numpy.testing.assert_array_equal(
                long_array,
                encode_long_array(block_array, (palette_size - 1).bit_length()),
            )

        # Make sure some test are ran in case the data file failed to load or has a wrong format.
        self.assertTrue(test_ran)

    def test_encode_decode(self):
        for signed in (False, True):
            for dense in (False, True):
                for bits_per_entry in range(4, 65):
                    for size in set(
                        [2 ** p for p in range(1, 13)] + list(range(100, 5000, 100))
                    ):
                        if signed:
                            arr = numpy.random.randint(
                                -(2 ** 63), 2 ** 63, size, dtype=numpy.int64
                            )
                        else:
                            arr = numpy.random.randint(
                                0, 2 ** 64, size, dtype=numpy.uint64
                            )
                        arr >>= 64 - bits_per_entry

                        packed = encode_long_array(arr, bits_per_entry, dense)
                        arr2 = decode_long_array(
                            packed, len(arr), bits_per_entry, dense=dense, signed=signed
                        )
                        numpy.testing.assert_array_equal(
                            arr,
                            arr2,
                            f"Long array does not equal. Dense: {dense}, bits per entry: {bits_per_entry}, size: {size}",
                        )


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy
import random
import math

from amulet.api.partial_3d_array.util import (
    get_sliced_array_size,
    sanitise_slice,
    unsanitise_slice,
)
from amulet.api.partial_3d_array import UnboundedPartial3DArray, BoundedPartial3DArray


class PartialArrayTestCase(unittest.TestCase):
    def test_get_sliced_array_size(self):
        arr = numpy.arange(16)
        start_stop_values = list(range(-20, 20)) + [None]
        step_values = list(range(-5, 0)) + list(range(1, 6))

        for _ in range(50000):
            start = random.choice(start_stop_values)
            stop = random.choice(start_stop_values)
            step = random.choice(step_values)
            sub_arr = arr[slice(start, stop, step)]
            rsize = sub_arr.size
            size = get_sliced_array_size(start, stop, step, arr.size)
            if rsize != size:
                get_sliced_array_size(start, stop, step, arr.size)
            self.assertEqual(
                rsize, size, f"{start}, {stop}, {step} | {rsize} != {size} {sub_arr}"
            )

    def test_unsanitise_slice(self):
        arr = numpy.arange(16)
        start_stop_values = list(range(-20, 20)) + [None]
        step_values = list(range(-5, 0)) + list(range(1, 6))

        for _ in range(50000):
            start = random.choice(start_stop_values)
            stop = random.choice(start_stop_values)
            step = random.choice(step_values)
            sub_arr = arr[slice(start, stop, step)]

            start2, stop2, step2 = sanitise_slice(start, stop, step, arr.size)
            start3, stop3, step3 = unsanitise_slice(start2, stop2, step2, arr.size)

            self.assertTrue(
                numpy.array_equal(sub_arr, arr[slice(start3, stop3, step3)])
            )

    def test_access(self):
        section_size = 8
        section_count = 16
        partial = UnboundedPartial3DArray(
            numpy.uint32,
            0,
            (section_size, section_size, section_size),
            (0, section_count),
        )
        array = numpy.arange(
            (section_size ** 3) * section_count, dtype=partial.dtype
        ).reshape((section_size, section_size * section_count, section_size))
        for cy in range(section_count):
            partial.add_section(
                cy, array[:, cy * section_size : (cy + 1) * section_size, :]
            )

        # test __getitem__ shape
        self.assertEqual(partial.shape, (section_size, math.inf, section_size))
        self.assertIsInstance(partial[:, :, :], BoundedPartial3DArray)
        self.assertEqual(
            partial[:, :, :].shape,
            (section_size, section_size * section_count, section_size),
        )
        self.assertEqual(
            partial[:, -100:500, :].shape, (section_size, 600, section_size)
        )
        self.assertEqual(partial[:, 0, :].shape, (section_size, 1, section_size))
        self.assertEqual(partial[0, :, 0].shape, (1, section_size * section_count, 1))

        # test __getitem__ values
        for slices in (
            (0, 0, 0),
            (2, 30, 2),
            (-1, 30, -1),
        ):
            self.assertEqual(partial[slices], array[slices], str(slices))

        # test nested __getitem__ values
        for slices in (
            (slice(None), slice(None), slice(None)),
            (slice(1, -1), slice(1, 50), slice(1, -1)),
            (slice(-1, 1, -1), slice(50, 1, -1), slice(-1, 1, -1)),
            (slice(-1, 1, -2), slice(50, 1, -2), slice(-1, 1, -2)),
        ):
            self.assertEqual(partial[slices].shape, array[slices].shape)
            for slices2 in (
                (0, 0, 0),
                (-1, -1, -1),
            ):
                self.assertEqual(
                    partial[slices][slices2],
                    array[slices][slices2],
                    f"[{slices}][{slices2}]",
                )

        for slices in (
            (slice(None), slice(None), slice(None)),
            (slice(1, -1), slice(1, 50), slice(1, -1)),
            (slice(-1, 1, -1), slice(50, 1, -1), slice(-1, 1, -1)),
        ):
            self.assertEqual(partial[slices].shape, array[slices].shape)
            for slices2 in (
                (slice(None), slice(None), slice(None)),
                (slice(1, -1), slice(1, -1), slice(1, -1)),
                (slice(-1, 1, -1), slice(-1, 1, -1), slice(-1, 1, -1)),
            ):
                self.assertEqual(
                    partial[slices][slices2].shape,
                    array[slices][slices2].shape,
                    f"[{slices}][{slices2}]",
                )
                for slices3 in (
                    (0, 0, 0),
                    (-1, -1, -1),
                ):
                    self.assertEqual(
                        partial[slices][slices3],
                        array[slices][slices3],
                        f"[{slices}][{slices2}][{slices3}]",
                    )

    def test_set(self):
        section_size = 4
        section_count = 16
        partial = UnboundedPartial3DArray(
            numpy.uint32,
            0,
            (section_size, section_size, section_size),
            (0, section_count),
        )

        partial[:, :4, :] = 5
        self.assertTrue(numpy.all(numpy.asarray(partial[:, :4, :]) == 5))
        self.assertEqual(tuple(partial.sections), (0,))
        self.assertTrue(numpy.all(partial.get_section(0) == 5))
        partial[:, 2:6, :] = 10
        self.assertTrue(numpy.all(numpy.asarray(partial[:, :2, :]) == 5))
        self.assertTrue(numpy.all(numpy.asarray(partial[:, 2:6, :]) == 10))
        self.assertTrue(numpy.all(numpy.asarray(partial[:, 6:, :]) == 0))

        self.assertEqual(tuple(partial.sections), (0, 1))

        arange = numpy.arange(section_size ** 3 * 4).reshape(
            (section_size, section_size * 4, section_size)
        )
        partial[:, int(section_size * 4.5) : int(section_size * 8.5), :] = arange
        self.assertTrue(
            numpy.array_equal(
                partial[:, int(section_size * 4.5) : int(section_size * 8.5), :], arange
            )
        )

        partial[:, int(section_size * 8.5) : int(section_size * 12.5), :] = partial[
            :, int(section_size * 4.5) : int(section_size * 8.5), :
        ]
        self.assertTrue(
            numpy.array_equal(
                partial[:, int(section_size * 8.5) : int(section_size * 12.5), :],
                arange,
            )
        )
        self.assertTrue(
            numpy.array_equal(
                partial[:, int(section_size * 4.5) : int(section_size * 8.5), :],
                partial[:, int(section_size * 8.5) : int(section_size * 12.5), :],
            )
        )

    def test_eq(self):
        section_size = 4
        section_count = 16
        partial = UnboundedPartial3DArray(
            numpy.uint32,
            0,
            (section_size, section_size, section_size),
            (0, section_count),
        )
        bounded_partial = partial[:, :, :]
        self.assertTrue(numpy.all(bounded_partial == 0))
        self.assertTrue(numpy.all(0 == bounded_partial))
        self.assertFalse(numpy.all(bounded_partial == 10))
        self.assertFalse(numpy.all(10 == bounded_partial))
        self.assertFalse(numpy.any(bounded_partial == 10))
        self.assertFalse(numpy.any(10 == bounded_partial))

        arange = numpy.arange(section_size ** 3 * 4).reshape(
            (section_size, section_size * 4, section_size)
        )
        partial[:, int(section_size * 4.5) : int(section_size * 8.5), :] = arange

        self.assertTrue(
            numpy.all(
                arange
                == partial[:, int(section_size * 4.5) : int(section_size * 8.5), :]
            )
        )
        self.assertTrue(
            numpy.all(
                partial[:, int(section_size * 4.5) : int(section_size * 8.5), :]
                == arange
            )
        )

        self.assertFalse(
            numpy.all(
                arange
                == partial[:, int(section_size * 8.5) : int(section_size * 12.5), :]
            )
        )
        self.assertFalse(
            numpy.all(
                numpy.zeros((section_size, section_size * 4, section_size))
                == partial[:, int(section_size * 4.5) : int(section_size * 8.5), :]
            )
        )
        self.assertTrue(
            numpy.all(
                numpy.zeros((section_size, section_size * 4, section_size))
                == partial[:, int(section_size * 8.5) : int(section_size * 12.5), :]
            )
        )

    def test_eq_index(self):
        section_size = 4
        section_count = 16
        partial = UnboundedPartial3DArray(
            numpy.uint32,
            0,
            (section_size, section_size, section_size),
            (0, section_count),
        )
        bounded_partial = partial[:, :, :]
        self.assertTrue(numpy.all(bounded_partial == 0))

        arange = numpy.arange(section_size ** 3 * section_count).reshape(
            (section_size, section_size * section_count, section_size)
        )
        partial[:, :, :] = arange

        self.assertTrue(numpy.all(bounded_partial == arange))

        bounded_partial[numpy.asarray(bounded_partial) % 2 == 1] = 10
        arange[arange % 2 == 1] = 10

        self.assertTrue(numpy.all(bounded_partial == arange))

    def test_getitem_bool(self):
        array = numpy.arange(16 ** 4, dtype=numpy.uint32).reshape((16, 16 * 16, 16))
        array[:, 128:, :] = 0
        partial_array = UnboundedPartial3DArray(
            numpy.uint32,
            0,
            (16, 16, 16),
            (0, 16),
            {sy: array[:, sy * 16 : (sy + 1) * 16, :].copy() for sy in range(8)},
        )

        array_slice = array[:, 0:200, :]
        bounded_partial_array = partial_array[:, 0:200, :]

        self.assertTrue(numpy.array_equal(array_slice, bounded_partial_array))

        bool_array = numpy.full(bounded_partial_array.shape, True, bool)
        self.assertTrue(
            numpy.array_equal(
                array_slice[bool_array], bounded_partial_array[bool_array]
            )
        )

        bool_array = numpy.random.choice([True, False], bounded_partial_array.shape)
        self.assertTrue(
            numpy.array_equal(
                array_slice[bool_array], bounded_partial_array[bool_array]
            )
        )



import unittest
import time

from amulet.api.errors import ChunkLoadError
from amulet import load_level
from tests.data.util import create_temp_world, clean_temp_world
from tests.data import worlds_src


class WorldTestBaseCases:
    # Wrapped in another class, so it isn't executed on it's own.

    class WorldTestCase(unittest.TestCase):
        def _setUp(self, world_name):
            self._world_name = world_name
            self.world = load_level(create_temp_world(world_name))

        def tearDown(self):
            self.world.close()
            clean_temp_world(self._world_name)

        def test_chunk_speed(self):
            chunk_count = 0
            start_time = time.time()
            for dimension in self.world.dimensions:
                for cx, cz in self.world.all_chunk_coords(dimension):
                    try:
                        chunk = self.world.get_chunk(cx, cz, dimension)
                    except ChunkLoadError:
                        pass
                    else:
                        chunk_count += 1
            end_time = time.time()
            print((end_time - start_time) / chunk_count)


class AnvilWorldTestCase(WorldTestBaseCases.WorldTestCase):
    def setUp(self):
        self._setUp(worlds_src.java_vanilla_1_12_2)


class Anvil2WorldTestCase(WorldTestBaseCases.WorldTestCase):
    def setUp(self):
        self._setUp(worlds_src.java_vanilla_1_13)


# class BedrockWorldTestCase(WorldTestBaseCases.WorldTestCase):
#     def setUp(self):
#         self._setUp(worlds_src.bedrock_vanilla_1_16)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import numpy
from typing import Tuple, Dict
from amulet_nbt import TAG_Compound, TAG_Long_Array
from amulet import log
from amulet.api.chunk import Chunk
from amulet.utils.world_utils import decode_long_array, encode_long_array

from .anvil_1444 import (
    Anvil1444Interface,
)


class Anvil1466Interface(Anvil1444Interface):
    """
    Added multiple height maps. Now stored in a compound.
    """

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "C|V1")

    @staticmethod
    def minor_is_valid(key: int):
        return 1466 <= key < 1467

    def _decode_height(
        self, chunk: Chunk, compound: TAG_Compound, bounds: Tuple[int, int]
    ):
        heights = self.get_obj(compound, "Heightmaps", TAG_Compound)
        chunk.misc["height_mapC"] = h = {}
        for key, value in heights.items():
            if isinstance(value, TAG_Long_Array):
                try:
                    h[key] = (
                        decode_long_array(
                            value.value,
                            256,
                            (bounds[1] - bounds[0]).bit_length(),
                            dense=self.LongArrayDense,
                        ).reshape((16, 16))
                        + bounds[0]
                    )
                except Exception as e:
                    log.warning(e)

    def _encode_height(
        self, chunk: Chunk, level: TAG_Compound, bounds: Tuple[int, int]
    ):
        maps = [
            "WORLD_SURFACE_WG",
            "OCEAN_FLOOR_WG",
            "MOTION_BLOCKING",
            "MOTION_BLOCKING_NO_LEAVES",
            "OCEAN_FLOOR",
        ]
        if self._features["height_map"] == "C|V1":  # 1466
            maps = ("LIQUID", "SOLID", "LIGHT", "RAIN")
        elif self._features["height_map"] == "C|V2":  # 1484
            maps.append("LIGHT_BLOCKING")
        elif self._features["height_map"] == "C|V3":  # 1503
            maps.append("LIGHT_BLOCKING")
            maps.append("WORLD_SURFACE")
        elif self._features["height_map"] == "C|V4":  # 1908
            maps.append("WORLD_SURFACE")
        else:
            raise Exception
        heightmaps_temp: Dict[str, numpy.ndarray] = chunk.misc.get("height_mapC", {})
        heightmaps = TAG_Compound()
        for heightmap in maps:
            if (
                heightmap in heightmaps_temp
                and isinstance(heightmaps_temp[heightmap], numpy.ndarray)
                and heightmaps_temp[heightmap].size == 256
            ):
                heightmaps[heightmap] = TAG_Long_Array(
                    encode_long_array(
                        heightmaps_temp[heightmap].ravel() - bounds[0],
                        (bounds[1] - bounds[0]).bit_length(),
                        self.LongArrayDense,
                    )
                )
        level["Heightmaps"] = heightmaps


export = Anvil1466Interface



from __future__ import annotations

from .leveldb_18 import (
    LevelDB18Interface,
)


class LevelDB19Interface(LevelDB18Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 19)


export = LevelDB19Interface



from .leveldb_32 import (
    LevelDB32Interface as ParentInterface,
)


class LevelDB33Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 33)


export = LevelDB33Interface



import unittest
from typing import Type
import os

from amulet import load_format
from amulet.api.data_types import VersionNumberAny
from amulet.api.wrapper import FormatWrapper
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ObjectWriteError

from amulet.level.formats.anvil_world import AnvilFormat
from amulet.level.formats.leveldb_world import LevelDBFormat
from amulet.level.formats.construction import ConstructionFormatWrapper
from amulet.level.formats.mcstructure import MCStructureFormatWrapper
from amulet.level.formats.schematic import SchematicFormatWrapper

from tests.data.util import clean_temp_world, clean_path


class CreateWorldTestCase(unittest.TestCase):
    def _test_create(
        self,
        cls: Type[FormatWrapper],
        level_name: str,
        platform: str,
        version: VersionNumberAny,
    ):
        path = clean_temp_world(level_name)

        # create, initialise and save the level
        level = cls(path)
        if level.requires_selection:
            level.create_and_open(
                platform,
                version,
                SelectionGroup([SelectionBox((0, 0, 0), (1, 1, 1))]),
                overwrite=True,
            )
        else:
            level.create_and_open(platform, version, overwrite=True)

        self.assertTrue(level.is_open, "The level was not opened by create_and_open()")
        self.assertTrue(
            level.has_lock, "The lock was not acquired by create_and_open()"
        )

        platform_ = level.platform
        version_ = level.version
        dimension_selections = {dim: level.bounds(dim) for dim in level.dimensions}

        level.save()
        level.close()

        self.assertFalse(level.is_open, "The level was not closed by close()")
        self.assertFalse(level.has_lock, "The lock was not lost by close()")

        self.assertTrue(os.path.exists(level.path))

        # reopen the level
        level2 = load_format(path)
        # check that the class is the same
        self.assertIs(level.__class__, level2.__class__)
        level2.open()

        self.assertTrue(level2.is_open, "The level was not opened by open()")
        self.assertTrue(level2.has_lock, "The lock was not acquired by open()")

        # check that the platform and version are the same
        self.assertEqual(level2.platform, platform_)
        self.assertEqual(level2.version, version_)
        self.assertEqual(set(level2.dimensions), set(dimension_selections))
        for dim, selection in dimension_selections.items():
            self.assertEqual(level2.bounds(dim), selection)
        level2.close()

        self.assertFalse(level2.is_open, "The level was not closed by close()")
        self.assertFalse(level2.has_lock, "The lock was not lost by close()")

        self.assertTrue(os.path.exists(level.path))

        level = cls(path)
        with self.assertRaises(ObjectWriteError):
            if level.requires_selection:
                level.create_and_open(
                    platform,
                    version,
                    SelectionGroup([SelectionBox((0, 0, 0), (1, 1, 1))]),
                )
            else:
                level.create_and_open(platform, version)

        clean_path(path)

    def test_anvil(self):
        self._test_create(
            AnvilFormat,
            "anvil_world",
            "java",
            (1, 16, 0),
        )

    def test_bedrock(self):
        self._test_create(
            LevelDBFormat,
            "leveldb_world",
            "bedrock",
            (1, 16, 0),
        )

    def test_construction(self):
        self._test_create(
            ConstructionFormatWrapper,
            "java.construction",
            "java",
            (1, 16, 0),
        )
        self._test_create(
            ConstructionFormatWrapper,
            "bedrock.construction",
            "bedrock",
            (1, 16, 0),
        )

    def test_mcstructure(self):
        self._test_create(
            MCStructureFormatWrapper,
            "bedrock.mcstructure",
            "bedrock",
            (1, 16, 0),
        )

    def test_schematic(self):
        self._test_create(
            SchematicFormatWrapper,
            "java.schematic",
            "java",
            (1, 12, 2),
        )
        self._test_create(
            SchematicFormatWrapper,
            "bedrock.schematic",
            "bedrock",
            (1, 12, 0),
        )


if __name__ == "__main__":
    unittest.main()



from typing import List
from setuptools import setup, find_packages
from Cython.Build import cythonize
import glob
import pkg_resources
import versioneer
import numpy


PROJECT_PREFIX = "amulet"


def load_requirements(path: str) -> List[str]:
    requirements = []
    with open(path) as f:
        for line in f.readlines():
            line = line.strip()
            if line and line[0] != "#":
                if line.startswith("-r "):
                    requirements += load_requirements(line[3:])
                else:
                    requirements.append(str(pkg_resources.Requirement.parse(line)))
    return requirements


required_packages = load_requirements("requirements.txt")

ext = []
pyx_path = f"{PROJECT_PREFIX}/**/*.pyx"
if next(glob.iglob(pyx_path, recursive=True), None):
    # This throws an error if it does not match any files
    ext += cythonize(
        pyx_path,
        language_level=3,
        annotate=True,
    )

setup(
    version=versioneer.get_version(),
    install_requires=required_packages,
    packages=find_packages(),
    include_package_data=True,
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=ext,
    include_dirs=[numpy.get_include()],
)



from .leveldb_31 import (
    LevelDB31Interface as ParentInterface,
)


class LevelDB32Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 32)


export = LevelDB32Interface



from __future__ import annotations

from .leveldb_10 import (
    LevelDB10Interface,
)


class LevelDB11Interface(LevelDB10Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 11)


export = LevelDB11Interface



from .leveldb_34 import (
    LevelDB34Interface as ParentInterface,
)


class LevelDB35Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 35)


export = LevelDB35Interface



from .leveldb_36 import (
    LevelDB36Interface as ParentInterface,
)


class LevelDB37Interface(ParentInterface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 37)


export = LevelDB37Interface



from __future__ import annotations

import os
import struct
from typing import Tuple, Any, Dict, Generator, Optional, List, Union, Iterable
import time
import glob
import shutil
import json

import amulet_nbt as nbt
from amulet.api.player import Player, LOCAL_PLAYER
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.wrapper import WorldFormatWrapper, DefaultVersion, DefaultSelection
from amulet.utils.format_utils import check_all_exist, load_leveldat
from amulet.api.errors import (
    DimensionDoesNotExist,
    ObjectWriteError,
    ChunkLoadError,
    PlayerDoesNotExist,
)
from amulet.api.data_types import (
    ChunkCoordinates,
    VersionNumberInt,
    PlatformType,
    DimensionCoordinates,
    AnyNDArray,
    Dimension,
)
from .dimension import AnvilDimensionManager
from amulet.api import level as api_level
from amulet.level.interfaces.chunk.anvil.base_anvil_interface import BaseAnvilInterface
from .data_pack import DataPack, DataPackManager


InternalDimension = str
OVERWORLD = "minecraft:overworld"
THE_NETHER = "minecraft:the_nether"
THE_END = "minecraft:the_end"


class AnvilFormat(WorldFormatWrapper):
    """
    This FormatWrapper class exists to interface with the Java world format.
    """

    _platform: PlatformType
    _version: VersionNumberInt

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`AnvilFormat`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._platform = "java"
        self._root_tag: nbt.NBTFile = nbt.NBTFile()
        self._levels: Dict[InternalDimension, AnvilDimensionManager] = {}
        self._dimension_name_map: Dict[Dimension, InternalDimension] = {}
        self._mcc_support: Optional[bool] = None
        self._lock_time = None
        self._data_pack: Optional[DataPackManager] = None
        self._shallow_load()

    def _shallow_load(self):
        try:
            self._load_level_dat()
        except:
            pass

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        if os.path.isfile(os.path.join(self.path, "icon.png")):
            self._world_image_path = os.path.join(self.path, "icon.png")
        else:
            self._world_image_path = self._missing_world_icon
        self.root_tag = nbt.load(os.path.join(self.path, "level.dat"))

    @staticmethod
    def is_valid(path: str) -> bool:
        if not check_all_exist(path, "level.dat"):
            return False

        try:
            level_dat_root = load_leveldat(path)
        except:
            return False

        if "Data" not in level_dat_root:
            return False

        if "FML" in level_dat_root:
            return False

        return True

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (True, True)}

    @property
    def version(self) -> VersionNumberInt:
        """The data version number that the world was last opened in. eg 2578"""
        if self._version == DefaultVersion:
            self._version = self._get_version()
        return self._version

    def _get_version(self) -> VersionNumberInt:
        return (
            self.root_tag.get("Data", nbt.TAG_Compound())
            .get("DataVersion", nbt.TAG_Int(-1))
            .value
        )

    @property
    def root_tag(self) -> nbt.NBTFile:
        """The level.dat data for the level."""
        return self._root_tag

    @root_tag.setter
    def root_tag(self, root_tag: Union[nbt.NBTFile, nbt.TAG_Compound]):
        if isinstance(root_tag, nbt.TAG_Compound):
            self._root_tag = nbt.NBTFile(root_tag)
        elif isinstance(root_tag, nbt.NBTFile):
            self._root_tag = root_tag
        else:
            raise ValueError("root_tag must be a TAG_Compound or NBTFile")

    @property
    def level_name(self) -> str:
        return str(self.root_tag["Data"].get("LevelName", ""))

    @level_name.setter
    def level_name(self, value: str):
        self.root_tag["Data"]["LevelName"] = nbt.TAG_String(value)

    @property
    def last_played(self) -> int:
        return self.root_tag["Data"]["LastPlayed"].value

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java {self.root_tag["Data"]["Version"]["Name"].value}'
        except Exception:
            return f"Java Unknown Version"

    @property
    def data_pack(self) -> DataPackManager:
        if self._data_pack is None:
            packs = []
            if (
                "DataPacks" in self.root_tag["Data"]
                and isinstance(self.root_tag["Data"]["DataPacks"], nbt.TAG_Compound)
                and "Enabled" in self.root_tag["Data"]["DataPacks"]
                and isinstance(
                    self.root_tag["Data"]["DataPacks"]["Enabled"], nbt.TAG_List
                )
            ):
                for pack in self.root_tag["Data"]["DataPacks"]["Enabled"]:
                    if isinstance(pack, nbt.TAG_String):
                        pack_name: str = pack.value
                        if pack_name == "vanilla":
                            pass
                        elif pack_name.startswith("file/"):
                            path = os.path.join(self.path, "datapacks", pack_name[5:])
                            if DataPack.is_path_valid(path):
                                packs.append(DataPack(path))
            self._data_pack = DataPackManager(packs)
        return self._data_pack

    @property
    def dimensions(self) -> List[Dimension]:
        return list(self._dimension_name_map.keys())

    def _register_dimension(
        self,
        relative_dimension_path: InternalDimension,
        dimension_name: Optional[Dimension] = None,
    ):
        """
        Register a new dimension.

        :param relative_dimension_path: The relative path to the dimension directory from the world root. "" for the world root.
        :param dimension_name: The name of the dimension shown to the user
        """
        if dimension_name is None:
            dimension_name: Dimension = relative_dimension_path

        if relative_dimension_path:
            path = os.path.join(self.path, relative_dimension_path)
        else:
            path = self.path

        if (
            relative_dimension_path not in self._levels
            and dimension_name not in self._dimension_name_map
        ):
            self._levels[relative_dimension_path] = AnvilDimensionManager(
                path, mcc=self._mcc_support
            )
            self._dimension_name_map[dimension_name] = relative_dimension_path
            bounds = DefaultSelection
            if self.version >= 2709:  # This number might be smaller

                def get_recursive(obj: nbt.TAG_Compound, *keys):
                    if isinstance(obj, nbt.TAG_Compound) and keys:
                        key = keys[0]
                        keys = keys[1:]
                        if key in obj:
                            if keys:
                                return get_recursive(obj[key], *keys)
                            else:
                                return obj[key]

                dimension_type = get_recursive(
                    self.root_tag.value,
                    "Data",
                    "WorldGenSettings",
                    "dimensions",
                    dimension_name,
                    "type",
                )
                if isinstance(dimension_type, nbt.TAG_String):
                    # the settings are in the data pack
                    dimension_type: str = dimension_type.value
                    if ":" in dimension_type:
                        namespace, base_name = dimension_type.split(":", 1)
                        if (
                            self.version >= 2834
                            and namespace == "minecraft"
                            and base_name == "overworld"
                        ):
                            bounds = SelectionGroup(
                                SelectionBox(
                                    (-30_000_000, -64, -30_000_000),
                                    (30_000_000, 320, 30_000_000),
                                )
                            )
                        dimension_path = (
                            f"data/{namespace}/dimension_type/{base_name}.json"
                        )
                        if self.data_pack.has_file(dimension_path):
                            with self.data_pack.open(dimension_path) as d:
                                try:
                                    dimension_settings_json = json.load(d)
                                except json.JSONDecodeError:
                                    pass
                                else:
                                    if (
                                        "min_y" in dimension_settings_json
                                        and isinstance(
                                            dimension_settings_json["min_y"], int
                                        )
                                    ):
                                        min_y = dimension_settings_json["min_y"]
                                        if min_y % 16:
                                            min_y = 16 * (min_y // 16)
                                    else:
                                        min_y = 0
                                    if (
                                        "height" in dimension_settings_json
                                        and isinstance(
                                            dimension_settings_json["height"], int
                                        )
                                    ):
                                        height = dimension_settings_json["height"]
                                        if height % 16:
                                            height = -16 * (-height // 16)
                                    else:
                                        height = 256

                                    bounds = SelectionGroup(
                                        SelectionBox(
                                            (-30_000_000, min_y, -30_000_000),
                                            (30_000_000, min_y + height, 30_000_000),
                                        )
                                    )

                elif isinstance(dimension_type, nbt.TAG_Compound):
                    # the settings are here
                    dimension_settings = dimension_type
                    if "min_y" in dimension_settings and isinstance(
                        dimension_settings["min_y"], nbt.TAG_Int
                    ):
                        min_y = dimension_settings["min_y"].value
                        if min_y % 16:
                            min_y = 16 * (min_y // 16)
                    else:
                        min_y = 0
                    if "height" in dimension_settings and isinstance(
                        dimension_settings["height"], nbt.TAG_Int
                    ):
                        height = dimension_settings["height"].value
                        if height % 16:
                            height = -16 * (-height // 16)
                    else:
                        height = 256

                    bounds = SelectionGroup(
                        SelectionBox(
                            (-30_000_000, min_y, -30_000_000),
                            (30_000_000, min_y + height, 30_000_000),
                        )
                    )

            self._bounds[dimension_name] = bounds

    def _get_interface_key(
        self, raw_chunk_data: Optional[Any] = None
    ) -> Tuple[str, int]:
        if raw_chunk_data:
            return (
                self.platform,
                raw_chunk_data.get("DataVersion", nbt.TAG_Int(-1)).value,
            )
        else:
            return self.max_world_version

    def _decode(
        self,
        interface: BaseAnvilInterface,
        dimension: Dimension,
        cx: int,
        cz: int,
        raw_chunk_data: Any,
    ) -> Tuple[Chunk, AnyNDArray]:
        bounds = self.bounds(dimension).bounds
        return interface.decode(cx, cz, raw_chunk_data, (bounds[0][1], bounds[1][1]))

    def _encode(
        self,
        interface: BaseAnvilInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> Any:
        bounds = self.bounds(dimension).bounds
        return interface.encode(
            chunk, chunk_palette, self.max_world_version, (bounds[0][1], bounds[1][1])
        )

    def _reload_world(self):
        # reload the level.dat in case it has changed
        self._load_level_dat()

        # create the session.lock file (this has mostly been lifted from MCEdit)
        self._lock_time = int(time.time() * 1000)
        try:
            with open(os.path.join(self.path, "session.lock"), "wb") as f:
                f.write(struct.pack(">Q", self._lock_time))
                f.flush()
                os.fsync(f.fileno())
        except PermissionError as e:
            self._is_open = False
            self._has_lock = False
            raise PermissionError(
                f"Could not access session.lock. The world may be open somewhere else.\n{e}"
            )

        self._is_open = True
        self._has_lock = True

        # the real number might actually be lower
        self._mcc_support = self.version > 2203

        self._levels.clear()
        self._bounds.clear()

        # load all the levels
        self._register_dimension("", OVERWORLD)
        self._register_dimension("DIM-1", THE_NETHER)
        self._register_dimension("DIM1", THE_END)

        for dir_name in os.listdir(self.path):
            level_path = os.path.join(self.path, dir_name)
            if os.path.isdir(level_path) and dir_name.startswith("DIM"):
                if AnvilDimensionManager.level_regex.fullmatch(dir_name) is None:
                    continue
                self._register_dimension(dir_name)

        for dimension_path in glob.glob(
            os.path.join(self.path, "dimensions", "*", "*", "region")
        ):
            dimension_path_split = dimension_path.split(os.sep)
            dimension_name = f"{dimension_path_split[-3]}:{dimension_path_split[-2]}"
            self._register_dimension(
                os.path.dirname(os.path.relpath(dimension_path, self.path)),
                dimension_name,
            )

    def _open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if os.path.isdir(self.path):
            if overwrite:
                shutil.rmtree(self.path)
            else:
                raise ObjectWriteError(
                    f"A world already exists at the path {self.path}"
                )
        self._version = self.translation_manager.get_version(
            self.platform, self.version
        ).data_version

        self.root_tag = root = nbt.TAG_Compound()
        root["Data"] = data = nbt.TAG_Compound()
        data["version"] = nbt.TAG_Int(19133)
        data["DataVersion"] = nbt.TAG_Int(self._version)
        data["LastPlayed"] = nbt.TAG_Long(int(time.time() * 1000))
        data["LevelName"] = nbt.TAG_String("World Created By Amulet")

        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        self._reload_world()

    @property
    def has_lock(self) -> bool:
        if self._has_lock:
            if self._lock_time is None:
                # the world was created not opened
                return True
            try:
                with open(os.path.join(self.path, "session.lock"), "rb") as f:
                    return struct.unpack(">Q", f.read(8))[0] == self._lock_time
            except:
                return False
        return False

    def pre_save_operation(
        self, level: api_level.BaseLevel
    ) -> Generator[float, None, bool]:
        changed_chunks = list(level.chunks.changed_chunks())
        height = self._calculate_height(level, changed_chunks)
        try:
            while True:
                yield next(height) / 2
        except StopIteration as e:
            height_changed = e.value

        # light = self._calculate_light(level, changed_chunks)
        # try:
        #     while True:
        #         yield next(light) / 2
        # except StopIteration as e:
        #     light_changed = e.value

        return height_changed  # or light_changed

    @staticmethod
    def _calculate_height(
        level: api_level.BaseLevel, chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        chunk_count = len(chunks)
        # it looks like the game recalculates the height value if not defined.
        # Just delete the stored height values so that they do not get written back.
        # tested as of 1.12.2. This may not be true for older versions.
        changed = False
        for i, (dimension, cx, cz) in enumerate(chunks):
            try:
                chunk = level.get_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                changed_ = False
                changed_ |= chunk.misc.pop("height_mapC", None) is not None
                changed_ |= chunk.misc.pop("height_map256IA", None) is not None
                if changed_:
                    changed = True
                    chunk.changed = True
            yield i / chunk_count
        return changed

    @staticmethod
    def _calculate_light(
        level: api_level.BaseLevel, chunks: List[DimensionCoordinates]
    ) -> Generator[float, None, bool]:
        """Calculate the height values for chunks."""
        # this is needed for before 1.14
        chunk_count = len(chunks)
        changed = False
        if level.level_wrapper.version < 1934:
            # the version may be less than 1934 but is at least 1924
            # calculate the light values
            pass
            # TODO
        else:
            # the game will recalculate the light levels
            for i, (dimension, cx, cz) in enumerate(chunks):
                try:
                    chunk = level.get_chunk(cx, cz, dimension)
                except ChunkLoadError:
                    pass
                else:
                    changed_ = False
                    changed_ |= chunk.misc.pop("block_light", None) is not None
                    changed_ |= chunk.misc.pop("sky_light", None) is not None
                    if changed_:
                        changed = True
                        chunk.changed = True
                yield i / chunk_count
        return changed

    def _save(self):
        """Save the data back to the disk database"""
        os.makedirs(self.path, exist_ok=True)
        for level in self._levels.values():
            level.save()
        self.root_tag.save_to(os.path.join(self.path, "level.dat"))
        # TODO: save other world data

    def _close(self):
        """Close the disk database"""
        pass

    def unload(self):
        for level in self._levels.values():
            level.unload()

    def _has_dimension(self, dimension: Dimension):
        return (
            dimension in self._dimension_name_map
            and self._dimension_name_map[dimension] in self._levels
        )

    def _get_dimension(self, dimension: Dimension):
        self._verify_has_lock()
        if self._has_dimension(dimension):
            return self._levels[self._dimension_name_map[dimension]]
        else:
            raise DimensionDoesNotExist(dimension)

    def all_chunk_coords(self, dimension: Dimension) -> Iterable[ChunkCoordinates]:
        if self._has_dimension(dimension):
            yield from self._get_dimension(dimension).all_chunk_coords()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return self._has_dimension(dimension) and self._get_dimension(
            dimension
        ).has_chunk(cx, cz)

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """Delete a chunk from a given dimension"""
        if self._has_dimension(dimension):
            self._get_dimension(dimension).delete_chunk(cx, cz)

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        self._get_dimension(dimension).put_chunk_data(cx, cz, data)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Dimension
    ) -> nbt.NBTFile:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        return self._get_dimension(dimension).get_chunk_data(cx, cz)

    def all_player_ids(self) -> Iterable[str]:
        """
        Returns a generator of all player ids that are present in the level
        """
        for f in glob.iglob(os.path.join(self.path, "playerdata", "*.dat")):
            yield os.path.splitext(os.path.basename(f))[0]
        if self.has_player(LOCAL_PLAYER):
            yield LOCAL_PLAYER

    def has_player(self, player_id: str) -> bool:
        if player_id == LOCAL_PLAYER:
            return "Player" in self.root_tag["Data"]
        else:
            return os.path.isfile(
                os.path.join(self.path, "playerdata", f"{player_id}.dat")
            )

    def _load_player(self, player_id: str) -> Player:
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player will be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        player_nbt = self._get_raw_player_data(player_id)
        dimension = player_nbt["Dimension"]
        # TODO: rework this when there is better dimension support.
        if isinstance(dimension, nbt.TAG_Int):
            if -1 <= dimension <= 1:
                dimension_str = {-1: THE_NETHER, 0: OVERWORLD, 1: THE_END}[
                    dimension.value
                ]
            else:
                dimension_str = f"DIM{dimension}"
        elif isinstance(dimension, nbt.TAG_String):
            dimension_str = dimension.value
        else:
            dimension_str = OVERWORLD
        if dimension_str not in self._dimension_name_map:
            dimension_str = OVERWORLD
        return Player(
            player_id,
            dimension_str,
            tuple(map(lambda t: t.value, player_nbt["Pos"])),
            tuple(map(lambda t: t.value, player_nbt["Rotation"])),
        )

    def _get_raw_player_data(self, player_id: str) -> nbt.NBTFile:
        if player_id == LOCAL_PLAYER:
            if "Player" in self.root_tag["Data"]:
                return self.root_tag["Data"]["Player"]
            else:
                raise PlayerDoesNotExist("Local player doesn't exist")
        else:
            path = os.path.join(self.path, "playerdata", f"{player_id}.dat")
            if os.path.exists(path):
                return nbt.load(path)
            raise PlayerDoesNotExist(f"Player {player_id} does not exist")


if __name__ == "__main__":
    import sys

    world_path = sys.argv[1]
    world = AnvilDimensionManager(world_path)
    chunk_ = world.get_chunk_data(0, 0)
    print(chunk_)
    world.put_chunk_data(0, 0, chunk_)
    world.save()



from __future__ import annotations

from .leveldb_3 import (
    LevelDB3Interface,
)


class LevelDB4Interface(LevelDB3Interface):
    def __init__(self):
        super().__init__()

        self._set_feature("chunk_version", 4)


export = LevelDB4Interface



import copy
import unittest

from amulet_nbt import TAG_Compound
from amulet import load_format
from tests.data.util import WorldTemp, for_each_world, BaseWorldTest
from tests.data.worlds_src import levels


@for_each_world(globals(), levels)
class BaseTestDecodeEncode(BaseWorldTest, unittest.TestCase):
    def test_decode_encode(self):
        """Test decoding chunk data and then encoding it again to make sure it matches."""
        with WorldTemp(self.WorldPath) as world_temp:
            level = load_format(world_temp.temp_path)
            level.open()
            chunk_count = 0
            for dimension in level.dimensions:
                for cx, cz in level.all_chunk_coords(dimension):
                    raw_chunk_data_in = level._get_raw_chunk_data(cx, cz, dimension)
                    raw_chunk_data = copy.deepcopy(raw_chunk_data_in)
                    interface = level._get_interface(raw_chunk_data)

                    if (
                        world_temp.metadata["world_data"]["platform"] == "java"
                        and world_temp.metadata["world_data"]["origin"] == "vanilla"
                    ):
                        # store references to the data
                        level_tag = raw_chunk_data.get("Level", TAG_Compound())

                    # decode the raw chunk data
                    chunk, chunk_palette = level._decode(
                        interface, dimension, cx, cz, raw_chunk_data
                    )

                    # TODO: uncomment this when the last few things in the Java format are sorted
                    # if (
                    #     world_temp.metadata["world_data"]["platform"] == "java"
                    #     and world_temp.metadata["world_data"]["origin"] == "vanilla"
                    # ):
                    #     self.assertFalse(raw_chunk_data, msg=self.WorldPath)
                    #     self.assertFalse(level_tag, msg=self.WorldPath)

                    raw_chunk_data_out = level._encode(
                        interface, chunk, dimension, chunk_palette
                    )
                    if chunk_count > 100:
                        break
                    chunk_count += 1
                    # if raw_chunk_data_in != raw_chunk_data_out:
                    #     print("Difference")
                    #     print(raw_chunk_data_in.value.find_diff(raw_chunk_data_out.value))
                    # print(raw_chunk_data_in.to_snbt())
                    # print(raw_chunk_data_out.to_snbt())

                    # self.assertEqual(raw_chunk_data_in, raw_chunk_data_out)

                    raw_chunk_data2 = copy.deepcopy(raw_chunk_data_out)
                    level._decode(interface, dimension, cx, cz, raw_chunk_data2)
            level.close()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Tuple, TYPE_CHECKING

from amulet_nbt import TAG_Compound, TAG_Int

from .anvil_na import (
    AnvilNAInterface,
)

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk


class Anvil0Interface(AnvilNAInterface):
    """
    Added the DataVersion tag
    Note that this has not been tested before 1.12
    """

    @staticmethod
    def minor_is_valid(key: int):
        return 0 <= key < 1444

    def _decode_root(self, chunk: Chunk, root: TAG_Compound):
        self._decode_data_version(root)

    @staticmethod
    def _decode_data_version(compound: TAG_Compound):
        # all versioned data must get removed from data
        if "DataVersion" in compound.value:
            del compound["DataVersion"]

    def _encode_data_version(
        self, root: TAG_Compound, max_world_version: Tuple[str, int]
    ):
        root["DataVersion"] = TAG_Int(max_world_version[1])


export = Anvil0Interface
